// $Id: bitmap.h,v 1.1 1997/03/25 10:18:43 yotam Exp $
#ifndef bItmAp
#define bItmAp

#include <iostream.h>

class BitSet;

// Tailored Bitmap class for TfmPk. Since X run count is used
// more than Y run, it is treated asymettrically.
class Bitmap
{
 public:
   Bitmap();
   Bitmap(const Bitmap &);
   Bitmap(const Bitmap &, unsigned int n, unsigned int d); // n/d magnify

   // The bitSetPArray is used when sharing BitSets. (e.g. repeat count TeX PK)
   Bitmap(unsigned int w, unsigned int h, const BitSet * const * bitSetPArray);

   virtual ~Bitmap();
   Bitmap & operator = (const Bitmap &rhs);

   // Bring to default constructor state
   void         clear();

   unsigned int value(unsigned int x, unsigned int y) const;
   void         set(unsigned int x, unsigned int y, unsigned int value = 1);

   // width, height growing upon sets.
   unsigned int w() const {return _w;};
   unsigned int h() const {return _h;};

   // For efficiency, tell the object we need such size
   void          bound(unsigned int w, unsigned int h);

   const BitSet  *bitSetArray() const { return _bitSet;};
 private:
   void           realloc(unsigned int newh);
   void           magnify(const Bitmap &, unsigned int n, unsigned int d);
   unsigned int   _w,
                  _h;

   BitSet         *_bitSet; // [_h}
   unsigned int   _step;
}; // Bitmap

extern ostream & operator << (ostream &os, const Bitmap &bitmap);


// Get 'efficient' list of rectangles whose disjoint union
// generates a given bitmap. The bitmap must not be changed
// while the iterator is active.
class BitmapRectIterator
{
 public:
   BitmapRectIterator(const Bitmap *pbm);   
   BitmapRectIterator(const BitmapRectIterator &);   
   BitmapRectIterator & operator =(const BitmapRectIterator &);   

   virtual ~BitmapRectIterator();

   bool nextRect(unsigned int& x, unsigned int& y, 
                 unsigned int& w, unsigned int& h); 
 private:
   unsigned int         _x, _y;   // current scan pointer
   const Bitmap         *_pbm;
   Bitmap               _covered; // The already scanned rectangles.
}; // BitmapRectIterator

#endif /* bItmAp */
