// $Id: block.c,v 1.3 1997/07/01 05:37:45 yotam Exp $

#include <stdlib.h>
#include <xwin.h>
#include "block.h"
#include "debug.h"

////////////////////////////////////////////////////////////////////////
TfmpkProgBlock::TfmpkProgBlock():
   argc(0),
   argv(0),
   xwinfo(0),
   interp(0),
   pTFM(0),
   pPK(0),
   pTk(0),
   pixelFontPerScreenNominator(1),
   pixelFontPerScreenDenominator(1),
   // pixelFontPerScreenSquareThreshold(1),
   pixelZoomScreenPerFont(2),
   chfill(6),
   cvfill(6),
   thfill(32),
   tvfill(32),

   _pickedChar(0),
   _bitmapFocusDirty(true),
   _focusGrid(true),
   _dirtyBitmap(true),
   _kernBitmapDirty(true),
   _pad(3)
{
} // TfmpkProgBlock


////////////////////////////////////////////////////////////////////////
TfmpkProgBlock::~TfmpkProgBlock()
{
   deleteHide();
   for (int ai = argc;  ai--;)
   {
      if (argv[ai])
      {
         delete [] argv[ai];
      }
   }
   delete [] argv;
   argv = 0;

   
} // ~TfmpkProgBlock


////////////////////////////////////////////////////////////////////////
bool
TfmpkProgBlock::copyArgs(int ac, const char*  const* av)
{
   bool  ok;

   DASSERT(argc == 0 && argv == 0); // to be called only once
   argc = ac;
   argv = new  char*  [ac];
   ok = (argv != 0);
   for (int ai = (ok ? ac : 0);  ai--;)
   {
      argv[ai] = new char[strlen(av[ai]) + 1];
      if (argv[ai] != 0)
      {
         (void)strcpy(argv[ai], av[ai]);
      }
      else
      {
         ok = false;
      }
   }
   return(ok);
} // copyArgs


#if 0
////////////////////////////////////////////////////////////////////////
void
TfmpkProgBlock::setPixelRatios(const char*  const* argv)
{
   US16  n, d, t, z;
   n = (US16)atoi(argv[0]);
   d = (US16)atoi(argv[1]);
   t = (US16)atoi(argv[2]);
   z = (US16)atoi(argv[3]);

   DASSERT(n && d && t && z); // all > 0
   pixelFontPerScreenNominator       = n;
   pixelFontPerScreenDenominator     = d;
   pixelFontPerScreenSquareThreshold = t;
   pixelZoomScreenPerFont            = z;
} // setPixelRatios
#endif


////////////////////////////////////////////////////////////////////////
bool
TfmpkProgBlock::setPreference(int argc, const char*  const*  const argv)
{
   bool ok = true;
   DPRNL("setPreference");
#if DEBUG
   for (int i = 0;  i < argc;  i++) {
      DPRNL("argv[" << i << "] = " << argv[i]); }
#endif
   if (!strcmp(argv[0], "color"))
   {
      ok = (argc == 3);
      if (ok)
      {
         ok = xwinfo->setColorGC(argv[1], argv[2]);
      }
   }
   else if (!strcmp(argv[0], "pixel"))
   {
      int n = -1; // init to silent warning
      ok = (argc == 3) && ((n = atoi(argv[2])) > 0);
      if (ok)
      {
         if (!strcmp(argv[1], "fpsNom"))
         {
            pixelFontPerScreenNominator = n;
            _dirtyBitmap = true;
         }
         else if (!strcmp(argv[1], "fpsDenom"))
         {
            pixelFontPerScreenDenominator = n;
            _dirtyBitmap = true;
         }
         else if (!strcmp(argv[1], "spfZoom"))
         {
            pixelZoomScreenPerFont = n;
            _bitmapFocusDirty = true;
            _kernBitmapDirty = true;
         }
         else
         {
            ok = false;
         }
      }
   }
   else
   {
      ok = false;
   }
   return(ok);           
} // setPreference


unsigned
TfmpkProgBlock::pixelTableMult(unsigned x) const
{
   const unsigned&  n = pixelFontPerScreenNominator;
   const unsigned&  d = pixelFontPerScreenDenominator;
   unsigned rc = (n * x + d - 1) / d;
   return(rc);
} // pixelTableMult
