// $Id: exten.c,v 1.1 1997/03/25 10:18:43 yotam Exp $
// function callback by the ximple widget drawing extensible character.

#include <stdlib.h>
#include <strstream.h>

#include <X11/Xlib.h>

#include "bignum.h"
#include "block.h"
#include "drawBitmap.h"
#include "tkinfo.h"
#include "texpk.h"
#include "textfm.h"
#include "xwin.h"
#include "debug.h"


class DrawState
{
 public:
   DrawState(const TeXtfm*   ptfm,
             const TeXpk*    ppk,
             Display*        d, 
             Window          w, 
             GC              g, 
             unsigned        m, 
             unsigned        x0,
             unsigned        y0
            ) :
            pTFM(ptfm),
            pPK(ppk),
            dpy(d),
            win(w),
            gc(g),
            mult(m),
            x(x0),
            y(y0)
            {};
   virtual ~DrawState() {};
   void     drawDown(US8 c, const Bitmap& bm);

   const TeXtfm*   pTFM;
   const TeXpk*    pPK;
   Display*        dpy;  
   const Window    win;
   const GC        gc;
      
   const unsigned  mult;   
   unsigned        x;
   unsigned        y;
}; // DrawState


////////////////////////////////////////////////////////////////////////
bool
TfmpkProgBlock::extenShow(int /* argc */, const char * const * const argv)
{
   bool  rc = true;
   extenIndex = atoi(argv[0]);
   extenKVal = atoi(argv[1]);
   _extenBitmapDirty = true;
   return(rc);
} // extenShow


////////////////////////////////////////////////////////////////////////
void
TfmpkProgBlock::extenShowBitmapUpdate()
{
   if (_extenBitmapDirty)
   {
      _extenBitmapDirty = false;
      TeXtfm::ExtensibleRecipe  extenRecipe = pTFM->exten(extenIndex);
      int               i;
      for (i = 4;  i--;)
      {

         US8  c = extenRecipe.byte(i);
         if ((c != 0) || (i == (int)TeXtfm::ExtensibleRecipe::rep))
         {
            Bitmap  asisBitmap(pPK->bitmapWidth(c), pPK->bitmapHeight(c),
                               pPK->bitmapBitSetArray(c));
            _extenTMBRBitmap[i] = Bitmap(asisBitmap, pixelZoomScreenPerFont, 1);
         }
         else
         {
            _extenTMBRBitmap[i] = Bitmap(); // empty
         }
      }
   }
} // extenShowBitmapUpdate


////////////////////////////////////////////////////////////////////////
void
TfmpkProgBlock::extenShowDraw()
{
   if (pPK && pTFM)
   {
      Tk_Window  tkwin = pTk->extenWin;
      DrawState   drawState(pTFM, pPK,
                            Tk_Display(tkwin), Tk_WindowId(tkwin),
                            xwinfo->gc(XWinInfo::Raster),
                            pixelZoomScreenPerFont,
                            _pad, _pad);

      {
         GC       gcClear = xwinfo->gc(XWinInfo::Clear);
         XFillRectangle(drawState.dpy, drawState.win, gcClear, 
                        0, 0, Tk_Width(tkwin), Tk_Height(tkwin));
      }


      typedef TeXtfm::ExtensibleRecipe ER;
      TeXtfm::ExtensibleRecipe  extenRecipe = pTFM->exten(extenIndex);

      US8  c[4], cc, cm;
      int  i;
      for (i = 4;  i--;  c[i] = extenRecipe.byte(i));

      if ((cc = c[ER::top]) != 0)
      {
         drawState.drawDown(cc, _extenTMBRBitmap[ER::top]);
      }
      int repSeg;
      for (repSeg = (((cm = c[ER::mid]) != 0) ? 2 : 1);  repSeg--;)
      {
         cc = c[ER::rep];
         for (i = extenKVal;  i--;)
         {
            drawState.drawDown(cc, _extenTMBRBitmap[ER::rep]);
         }
         if (repSeg == 1)
         {
            drawState.drawDown(cm, _extenTMBRBitmap[ER::mid]);
         }
      }
      if ((cc = c[ER::bot]) != 0)
      {
         drawState.drawDown(cc, _extenTMBRBitmap[ER::bot]);
      }
         
   }
} // TfmpkProgBlock::extenShowDraw


////////////////////////////////////////////////////////////////////////
void
TkInfo::extenDraw(Tk_Window  DCODE(tkwin), ClientData appData)
{
   DPRNL("TkInfo::extenDraw");
   TfmpkProgBlock*  pb = (TfmpkProgBlock *)appData;
   DASSERT(tkwin == pb->pTk->extenWin);
   pb->extenShowDraw();
} // TkInfo::extenDraw


////////////////////////////////////////////////////////////////////////
bool
TfmpkProgBlock::extenShowWinSize()
{
   DPRNL("TkInfo::extenShowWinSize");
   bool  ok = false;
   focusBitmapUpdate();
   if (pPK && pPK->active() && pTFM)
   {
      ok = true;
      ostrstream  ostr(interp->result, TCL_RESULT_SIZE - 1);
      tableDims(); // ??
      extenShowBitmapUpdate();

      TeXtfm::ExtensibleRecipe  extenRecipe = pTFM->exten(extenIndex);

         DCODE(US8  T = extenRecipe.byte(TeXtfm::ExtensibleRecipe::top));
               US8  M = extenRecipe.byte(TeXtfm::ExtensibleRecipe::mid);
         DCODE(US8  B = extenRecipe.byte(TeXtfm::ExtensibleRecipe::bot));
               US8  R = extenRecipe.byte(TeXtfm::ExtensibleRecipe::rep);
      DPRNL(hex << "T=0x" << (int)T << ", M=0x" << (int)M << 
                   ", B=0x" << (int)B << ", R=0x" << (int)R << dec); 
         
      unsigned  w = pTFM->cWidth(R), 
                ch, h = 0;
      int       i;
      for (i = 3;  i--;)
      {
         DASSERT(i != TeXtfm::ExtensibleRecipe::rep);
         US8       c = extenRecipe.byte(i);
         ch = (c ? pTFM->cHeight(c) + pTFM->cDepth(c) : 0);
         h += ch;
         DPRNL("i=" << i << ", ch=" << ch << ", h=" << h);
      }
      
      ch = pTFM->cHeight(R) + pTFM->cDepth(R);
      h += ((M != 0 ? 2 : 1) * extenKVal * ch);  // (R^k) or  (R^k)M(R^k)
      DPRNL("TFM: ch=" << ch << ", w=" << w << ", h=" << h);

      US32  ds = pTFM->hDesignSize();
      US32  hppp = pPK->hppp();
      US32  vppp = pPK->vppp();
      w = n1n2n3_div2p(hppp, ds, w, 16 + 20 + 20);
      h = n1n2n3_div2p(vppp, ds, h, 16 + 20 + 20);
      DPRNL("PK: " << ", w=" << w << ", h=" << h);
      
      w *= pixelZoomScreenPerFont;
      h *= pixelZoomScreenPerFont;
      DPRNL("w=" << w << ", h=" << h << ", p2s=" << pixelZoomScreenPerFont);
      ostr << w << ' ' << h << ends;
      
   }
   return(ok);
} // TkInfo::extenShowWinSize




////////////////////////////////////////////////////////////////////////
int
TkInfo::extenConfigure(Tk_Window, ClientData /* appData */)
{
   DPRNL("TkInfo::extenConfigure");
   int   rc = TCL_OK;
   return(rc);
} // TkInfo::extenConfigure


////////////////////////////////////////////////////////////////////////
void
DrawState::drawDown(US8 c, const Bitmap& bm)
{
   Int32     hoff = pPK->hoff(c);
   Int32     voff = pPK->voff(c);
   unsigned  xoff = x + mult * (-hoff);
   unsigned  yoff = y + mult * (-voff);

   drawBitmap(dpy, win, gc, &bm, xoff, yoff);

   US32  ds = pTFM->hDesignSize();
   Int32  hd = pTFM->cHeight(c) + pTFM->cDepth(c);
   US32  vppp = pPK->vppp();
   unsigned  h = n1n2n3n4_div2p(vppp, ds, hd, mult, 16 + 20 + 20);
   y += h;
} // drawDown

