#include <string>
#include <string.h>
#include "baseint.h"
#include "debug.h"

////////////////////////////////////////////////////////////////////////
static unsigned
multdiv(unsigned n, unsigned d, unsigned x)
{
   unsigned  md = (x * n + d - 1) / d;
   return(md);
} // multdiv


////////////////////////////////////////////////////////////////////////
static unsigned
denom2nomAndBadness(unsigned& nom, unsigned sI, unsigned sM, unsigned denom)
{
   // we need to find maximal 'nom' such that  
   //  (nom * sI) / denom <= sM
   nom = (sM * denom) / sI;
   unsigned  baddness = sM - (nom * sI) / denom;
   return(baddness);
} // denom2nomAndBadness


////////////////////////////////////////////////////////////////////////
// denom - I/O
static bool 
fitzoom(US16& nom, US16& denom, unsigned  sI, unsigned  sM)
{
   bool  ok = true;
   unsigned  minDenom = (sI + sM - 1) / sM;

   unsigned  maxDenom = (denom < 4 ? 4 : denom);
   if (maxDenom < minDenom)
   {
      maxDenom = minDenom;
   }
   // find the best denom
   unsigned  leastBadness = sM;
   for (unsigned d = minDenom;  d <= maxDenom;  d++)
   {
      unsigned  n, baddness = denom2nomAndBadness(n, sI, sM, d);
      if (leastBadness > baddness)
      {
         leastBadness = baddness;
         nom = n;
         denom = d;
      }
   }

#if 0
   unsigned  n = nom;
   unsigned  d = denom;
   for ( ;  (n > 1) && (multdiv(n, d, sI) > sM);  n--);
   for ( ;  (multdiv(n, d, sI) > sM);  d++);
   for ( ;  (n > 1) && (multdiv(n, d, sI) > sM);  n--);
   nom = n;
   denom = d;
#endif
   return(ok);
} // fitzoom


////////////////////////////////////////////////////////////////////////
bool
fitzoom
(
   // output
   unsigned&  xZoom,
   unsigned&  yZoom,
   US16&      zoomNom,
   US16&      zoomDenom,

   unsigned   xIdeal,
   unsigned   yIdeal,
   unsigned   xMax,
   unsigned   yMax,
   US16       nom,
   US16       denom
)
{
   bool  ok = true;
   xZoom = multdiv(nom, denom, xIdeal);
   yZoom = multdiv(nom, denom, yIdeal);
   if ((xZoom > xMax) || (yZoom > yMax))
   {
      // which is the bigger ratio?
      if (xZoom * yMax > yZoom * xMax)
      {
         ok = fitzoom(nom, denom, xIdeal, xMax);
      }
      else
      {
         ok = fitzoom(nom, denom, yIdeal, yMax);
      }
      xZoom = multdiv(nom, denom, xIdeal);
      yZoom = multdiv(nom, denom, yIdeal);
   }
   zoomNom   = nom;
   zoomDenom = denom;
   return(ok);
} // fitzoom


#if !defined(MAIN)
#include <stdlib.h>
#include <strstream.h>
#include "block.h"
#include "texpk.h"

////////////////////////////////////////////////////////////////////////
bool
TfmpkProgBlock::fitZoomTable(const char* const* argv)
{
   bool         ok = true;
   unsigned     xMax, yMax;
   const char*  maxGeo = argv[0];
   US16         hintNom = (US16)atoi(argv[1]);
   US16         hintDenom = (US16)atoi(argv[2]);
   // libg++ prior to 2.7.2 leaks memory for istrstream
   // char         xSep;
   // {istrstream(maxGeo) >> xMax >> xSep >> yMax;}
   const char*  px = strchr(maxGeo, 'x');  DASSERT(px != 0);
   xMax = atoi(maxGeo);
   yMax = (px ? atoi(px + 1) : 0);
   if (ok && (xMax != 0) || (yMax != 0) && (pPK != 0) && pPK->active())
   {
     
      US16   minRow = charFirst() / 16, 
             maxRow = charLast() / 16;   

      if (minRow <= maxRow)
      {
         unsigned  xUnused, yUnused;
         unsigned  xI = 16 * (maxRef(Left) + maxRef(Right));
         unsigned  yI = (maxRow - minRow + 1) * (maxRef(Up) + maxRef(Down));
         fitzoom(xUnused, yUnused,
                 pixelFontPerScreenNominator, pixelFontPerScreenDenominator,
                 xI, yI, xMax, yMax, hintNom, hintDenom);
      }
   }
   tableDims();
   return(ok);
} // fitzoom


#endif


#if MAIN
////////////////////////////////////////////////////////////////////////
//                               Test
#include <iostream.h>

int
main(int, char**)
{
   unsigned   xZoom,
              yZoom,   
              zoomNom,
              zoomDenom,
              
              xIdeal,
              yIdeal,
              xMax,
              yMax,
              nom,
              denom;

   bool  ok = true;
   while (ok && !cin.fail())   
   {
      cin >> xIdeal >> yIdeal >> xMax >> yMax >> nom >> denom;
      ok = fitzoom(xZoom, yZoom, zoomNom, zoomDenom,
                   xIdeal, yIdeal, xMax, yMax, nom, denom);
      cout << "xIdeal=" << xIdeal << ", yIdeal=" << yIdeal << 
              ", xMax=" << xMax << ", yMax=" << yMax << 
              ", nom=" << nom << ", denom=" << denom << endl;
      cout << "=> xZoom=" << xZoom << ", yZoom=" << yZoom << 
               ", zoomNom=" << zoomNom << ", zoomDenom=" << zoomDenom << endl;
   }
} // main
#endif
