// $Id: kern.c,v 1.2 1997/03/25 10:18:43 yotam Exp $
// function callback by the ximple widget drawing kerned pair of characters.

#include <stdlib.h>
#include <strstream.h>
#include "bignum.h"
#include "block.h"
#include "drawBitmap.h"
#include "tkinfo.h"
#include "texpk.h"
#include "textfm.h"
#include "xwin.h"
#include "debug.h"


////////////////////////////////////////////////////////////////////////
bool
TfmpkProgBlock::kernShow(int /* argc */, const char * const * const argv)
{
   bool  rc = true;
   _kernShowIndex      = (US32)atol(argv[0]);
   _kernShowOwnerIndex = (US32)atol(argv[1]);
   _kernShowKernOn     = (bool)atoi(argv[2]);
   DPRNL("kernShow: " << _kernShowIndex << ' ' << _kernShowOwnerIndex << ' ' <<
         _kernShowKernOn);
   _kernBitmapDirty = true;
   return(rc);
} // kernShow


////////////////////////////////////////////////////////////////////////
static void
kernChars(US8* kc2, const TeXtfm& tfm, unsigned ki, unsigned oi)
{
   TeXtfm::LigKern  lk = tfm.lig_kern(ki);
   DASSERT(lk.kern_flag());
   kc2[1] = lk.next();
   US8     owners[256]; // must be enough
   (void)tfm.ligKernOwners(owners, oi + 1, ki);
   kc2[0] = owners[oi];
} // kernChars


////////////////////////////////////////////////////////////////////////
void
TfmpkProgBlock::kernShowBitmapUpdate()
{
   DPRNL("kernShowBitmapUpdate");
   if (_kernBitmapDirty)
   {
      _kernBitmapDirty = false;
      if (pTFM && pPK)
      {
         US8  kc[2];
         kernChars(kc, *pTFM, _kernShowIndex, _kernShowOwnerIndex);
         int  i;
         for (i = 2;  i--;)
         {
            US8  c = kc[i];
            DPRNL("kc[" << i << "] = 0x" << hex << (unsigned)c << dec);
            Bitmap  asisBitmap(pPK->bitmapWidth(c), pPK->bitmapHeight(c),
                             pPK->bitmapBitSetArray(c));
            _kernBitmap[i] = Bitmap(asisBitmap, pixelZoomScreenPerFont, 1);
            DPR0("i=" << i << ", c=0x" << hex << (unsigned)c << endl <<
                  "asis=" << endl << asisBitmap << endl <<
                  "zoomed=" <<endl << _kernBitmap[i]);
         }
      }
   }
} // kernShowBitmapUpdate


////////////////////////////////////////////////////////////////////////
void
TfmpkProgBlock::kernShowDraw()
{
   DPRNL("TfmpkProgBlock::kernShowDraw");
   if (pPK && pTFM)
   {
      Tk_Window  tkwin = pTk->kernWin;
      // Tk_GeometryRequest(tkwin, _bitmapFocus.w(), _bitmapFocus.h());
      Display*  dpy = Tk_Display(tkwin);
      Window    win = Tk_WindowId(tkwin);
      unsigned mult = pixelZoomScreenPerFont;

      {
         GC       gcClear = xwinfo->gc(XWinInfo::Clear);
         XFillRectangle(dpy, win, gcClear, 
                        0, 0, Tk_Width(tkwin), Tk_Height(tkwin));
      }

      // see if any kern
      unsigned   lki, nl = pTFM->length(TeXtfm::nl);    
      bool anyKern = false;
      for (lki = 0;  (lki < nl) && !anyKern;  lki++)
      {
         anyKern = pTFM->lig_kern(lki).kern_flag();
         if ((_kernShowIndex < 0) && anyKern)
         {
            _kernShowIndex = lki; // default: first kern to show
         }
      }
      if (anyKern)
      {
         kernShowBitmapUpdate();
         US8            kc[2], c;
         const TeXtfm&  tfm = *pTFM;  // most values are Fixed-Point 2^20
         kernChars(kc, tfm, _kernShowIndex, _kernShowOwnerIndex);
         GC             gcRaster = xwinfo->gc(XWinInfo::Raster);

         US32  ds = pTFM->hDesignSize();
         US32  hppp = pPK->hppp();
         TeXtfm::LigKern  lk = tfm.lig_kern(_kernShowIndex);
         unsigned         ki = lk.reminder();
         Int32  k = tfm.kern(ki);

         c = kc[0];
         Int32  cw = (Int32)tfm.cWidth(c);
         Int32  px0, x0 = -( cw + k);
         if (x0 < 0)
         {
            px0 = x0 = 0;
         }
         else
         {
            px0 = n1n2n3n4_div2p(hppp, ds, x0, mult, 16 + 20 + 20);
         }

         Int32     hoff = pPK->hoff(c);
         Int32     voff = pPK->voff(c);
         unsigned  xoff = _pad + mult * (maxRef(Left) - hoff) + px0;
         unsigned  yoff = _pad + mult * (maxRef(Up) - voff);

         drawBitmap(dpy, win, gcRaster, &_kernBitmap[0], xoff, yoff);

         x0 += cw;
         if (_kernShowKernOn)
         {   
            x0 += k;
         }
         px0 = n1n2n3n4_div2p(hppp, ds, x0, mult, 16 + 20 + 20);

         c = kc[1];
         hoff = pPK->hoff(c);
         voff = pPK->voff(c);
         xoff = _pad + mult * (maxRef(Left) - hoff) + px0;
         yoff = _pad + mult * (maxRef(Up)   - voff);

         drawBitmap(dpy, win, gcRaster, &_kernBitmap[1], xoff, yoff);

      }
   }
} // TfmpkProgBlock::kernShowDraw


////////////////////////////////////////////////////////////////////////
void
TkInfo::kernDraw(Tk_Window DCODE(tkwin), ClientData appData)
{
   DPRNL("TkInfo::kernDraw");
   TfmpkProgBlock*  pb = (TfmpkProgBlock *)appData;
   DASSERT(tkwin == pb->pTk->kernWin);
   pb->kernShowDraw();
} // TkInfo::kernDraw


////////////////////////////////////////////////////////////////////////
bool
TfmpkProgBlock::kernShowWinSize()
{
   DPRNL("TkInfo::kernShowWinSize");
   bool  ok = false;
   focusBitmapUpdate();
   if (pPK && pPK->active() && pTFM)
   {
      ok = true;
      ostrstream  ostr(interp->result, TCL_RESULT_SIZE - 1);
      // ostr << _bitmapFocus.w() << ' ' <<  _bitmapFocus.h() << ends;
      tableDims();
      // kernShowBitmapUpdate();

      // The maximal BBox for the bitmap
      unsigned  w  = maxRef(Left) + maxRef(Right);
      unsigned  h  = maxRef(Up) + maxRef(Down);
      DPRNL("maxBB: " << w << " * " << h);
      w *= 2; // two charactares are to be drawn

      // find the maximal positive kern value
      int  maxKern = 0, ki;
      for  (ki = pTFM->length(TeXtfm::nk) - 1;  ki > 0;  ki--)
      {
         int  k =  pTFM->kern(ki);
         if (maxKern < k)
         {
            maxKern = k;
         }
      }

      US32  ds = pTFM->hDesignSize();
      US32  hppp = pPK->hppp();
      w *= pixelZoomScreenPerFont;
      h *= pixelZoomScreenPerFont;

      unsigned  kernMaxPixel = n1n2n3n4_div2p(hppp, ds,  maxKern, 
                                              pixelZoomScreenPerFont,
                                              16 + 20 + 20);
      DPRNL("kernMaxPixel=" << kernMaxPixel);
      w += kernMaxPixel;
      ostr << w << ' ' << h << ends;
   }
   return(ok);
} // TkInfo::kernShowWinSize




////////////////////////////////////////////////////////////////////////
int
TkInfo::kernConfigure(Tk_Window, ClientData /* appData */)
{
   DPRNL("TkInfo::kernConfigure");
   int   rc = TCL_OK;
   return(rc);
} // TkInfo::kernConfigure




