// $Id: main.c,v 1.7 1997/07/25 05:44:30 yotam Exp yotam $

# include <string>
# include <ctype.h>
# include <stdlib.h>
# include <string.h>
# include <unistd.h>
# include <sys/types.h>
# include <sys/stat.h>

#include <iostream.h>
#include <strstream.h>
#include <tk.h>

#include "baseint.h"
#include "block.h"
#include "lspath.h"
#include "tkinfo.h"
#include "ximple.h"
#include "debug.h"
#if MEMCHECK
# include "newdel.h"
#endif

#if !defined(VERSION)
# define VERSION  "Unknown"
#endif



////////////////////////////////////////////////////////////////////////
static int
tfmpkXError(Display* dpy, XErrorEvent* ev)
{
   char buffer[1024];
   XGetErrorText(dpy, ev->error_code, buffer, sizeof(buffer));
   cerr << "XError in tfmpk:" << endl << buffer << endl;
   DASSERT(false);
   return 0;
} // tfmpkXError


////////////////////////////////////////////////////////////////////////
static int
tfmpkCmd
(
   ClientData  clientData, 
   Tcl_Interp* DCODE(interp), 
   int         argc, 
   char**      argv
)
{
   int  rc = TCL_OK;

#if DEBUG
   DASSERT(argc >= 2 && !strcmp(argv[0], "tfmpk"));
   DPR0("clientData=" << clientData << ", interp=" << interp << 
           ", argc=" << argc << endl);
   for (int  i = 0;  i < argc;  i++)
   {
      DPR0("argv[" << i << "]=" << argv[i] << endl);
   }
#endif

   TfmpkProgBlock*  pb = (TfmpkProgBlock*)clientData;
   DASSERT(pb->interp == interp);
   rc = pb->cmd(argc, argv);

   return(rc);   
} // tfmpkCmd


////////////////////////////////////////////////////////////////////////
static char* 
strClone(const char* s)
{
   unsigned  l = strlen(s);
   char*     ns = new char[l + 1];
   (void)strcpy(ns, s);
   return(ns);
} // strClone


////////////////////////////////////////////////////////////////////////
static void
showTclError(Tcl_Interp*  interp)
{
   const char* errInfo = Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY);
   cerr << "Tcl Error:" << endl << 
     "result: " << interp->result << endl <<
     "errorInfo: " << errInfo << endl << 
     "errorLine=" << interp->errorLine << endl;
} // showTclError


////////////////////////////////////////////////////////////////////////
static bool
pathFindEval
(
   Tcl_Interp*                interp,
   unsigned                   nd, 
   const char* const*  const  dirs, 
   const char*                script
)
{
   bool         ok = true;
   struct stat  dumStat;
   bool         found = false;
   string       fullPath;
   unsigned     di;
   for (di = 0;  ok && !found && di < nd;  di++)   
   {
      unsigned  ld = strlen(dirs[di]);
      if (ld > 0)
      {
         static const string slash("/");
         fullPath = string(dirs[di]) + slash + string(script);
         DPRNL("Trying fullPath=" << fullPath.c_str());
      }
      const char*  ccandid = (ld > 0 ? fullPath.c_str() : script);
      char*        scandid = strClone(ccandid);
      if (stat(scandid, &dumStat) == 0)
      {
         found = true;  DPRNL("found " << scandid);
         ok = (Tcl_EvalFile(interp, scandid) == TCL_OK);
         if (!ok)
         {
            cerr << "Failed to evaluate " << scandid << endl;
            showTclError(interp);
         }
      }
      delete [] scandid;
   }
   return(ok);
} // pathFindEval


#define SZA(a) (sizeof(a) / sizeof(a[0]))

////////////////////////////////////////////////////////////////////////
static bool
initScripts(Tcl_Interp* interp, int argc, const char*  const*  const argv)
{
   static const char* envNames[] = {"PATH", "TCL_LIBRARY", "HOME"};
   static const char* cenvdirs[SZA(envNames)];
   static       char* envdirs[SZA(envNames)];

   bool  ok = true;

   // copy paths and count colons within paths to get # of directories.
   int  i, nd = 2; // 2  for current + exec-dircetory
   for (i = SZA(envNames);  i--;)
   {
      envdirs[i] = 0;
      cenvdirs[i] = getenv(envNames[i]);
      const char* envdir = cenvdirs[i];
      if (envdir)
      {
         nd++;
         const char*  colon;
         for (colon = envdir - 1; (colon = strchr(colon + 1, ':')) != 0;  nd++);
         envdirs[i] = strClone(envdir);
      }
   }

   char**  dirs = new char* [nd];

   dirs[0] = ""; // current directory
   nd = 1;

   // get directory of executable.
   const char* argv0 = argv[0];
   char*  execdir = 0;
   const char*  slash = strrchr(argv0, '/');
   if (slash)
   {
      int  l = slash - argv0;
      dirs[1] = execdir = new char[l + 1];
      strncpy(execdir, argv0, l);
      execdir[l] = '\0';
      nd = 2;
   }
   // else we search PATH anyway.

   // Split the paths, by nullifying colons. Set dirs to directory positions.
   for (i = SZA(envNames);  i--;)
   {
      char*  pos = envdirs[i];
      if (pos)
      {
         char*  colon = dirs[nd++] = pos;
         while ((colon = strchr(pos, ':')) != 0)
         {
            *colon = '\0';
            pos = colon + 1;
            dirs[nd++] = pos;
         }
      }
   }

   ok = ok && pathFindEval(interp, nd, dirs, "tfmpk.tcl");

   if (ok)
   {
      ostrstream  tfmpkMain;
      tfmpkMain << "tfmpkMain";
      for (i = 0;  i < argc;  i++)
      {
         tfmpkMain << ' ' << argv[i];
      }
      tfmpkMain << ends;
      char*   strTfmpkMain = strClone( tfmpkMain.str() );
      DPRNL("Calling Tcl_Eval(interp, " << strTfmpkMain << ")");
      (void)Tcl_SetVar2(interp, "tfmpkGlob","initscript", "", TCL_GLOBAL_ONLY);
      int  mainRC = Tcl_Eval(interp, strTfmpkMain);
      ok = (mainRC == TCL_OK);
      if (ok)
      {
         const char* initscript = Tcl_GetVar2(interp, "tfmpkGlob", "initscript",
                                              TCL_GLOBAL_ONLY);
         ok = (strcmp(initscript, "") == 0);
      }
      else
      {
         cerr << "Failed to Evaluate:" << endl << strTfmpkMain << endl;
         showTclError(interp);
      }
      delete [] strTfmpkMain;
#if defined(__GNUC__)
      tfmpkMain.freeze(0);
#endif
   }

   if (execdir) 
   {
      delete [] execdir;
   }
   for (i = SZA(envNames);  i--;)
   {
      if (envdirs[i])
      {
         delete [] envdirs[i];
      }
   }
   delete [] dirs;
   return(ok);
} // initScripts


////////////////////////////////////////////////////////////////////////
int
main1(int argc, char** argv)
{
#if MEMCHECK
   YMCHECKMEM memcheker;
#endif
   // DASSERT(false);  // just to hook a debugger
   bool  ok = true;
   TfmpkProgBlock  block;
   TkInfo          infoTk;
   block.pTk = &infoTk;

   DPRNL("pid=" << getpid());
   // Tk_Main(argc, argv, appInit); because appInit cannot get argv

   char*  basename = strrchr(argv[0], '/');
   basename = (basename ? basename + 1 : argv[0]);

   int  i; bool silent = false;
   for (i = 1;  i < argc && !silent;  silent = !strcmp(argv[i++], "-silent"));
   if (!silent)
   {
      cerr << basename << " Version " << VERSION << endl;
   }

   char* Basename = strClone(basename);
   Basename[0] = toupper(Basename[0]);
   Tcl_Interp*  interp = block.interp = Tcl_CreateInterp();

   if (Tcl_Init(interp) == TCL_ERROR ||
       Tk_Init(interp) == TCL_ERROR || 
       Ximple_Init(interp, infoTk.assignFuncs, &block, 0) == TCL_ERROR)
   {
      cerr << basename << ": failed to initialize Tcl-package(s)" << endl;
      ok = false;
   }
   else
   {
      (void)Tcl_SetVar2(interp, "tfmpkGlob","version", VERSION, 
                        TCL_GLOBAL_ONLY);
      Tk_Window  mainWindow = Tk_MainWindow(interp);
      (void)XSetErrorHandler(tfmpkXError);
      Tk_SetClass(mainWindow, Basename);
      Tcl_CreateCommand(interp, "tfmpk", tfmpkCmd, 
                        (ClientData)&block, 
                        (Tcl_CmdDeleteProc*)NULL);
      Tcl_CreateCommand(interp, "lspath", lspath, 
                        (ClientData)0, 
                        (Tcl_CmdDeleteProc*)NULL);

      Tk_MakeWindowExist(mainWindow); // so we can configure colors
      ok = initScripts(interp, argc, argv);
      if (ok)
      {
         Tk_MainLoop();
      }
   }
   
   delete [] Basename;
   Tcl_DeleteInterp(interp);
   return(!ok);
} // main1


////////////////////////////////////////////////////////////////////////
int
main(int argc, char** argv)
{
   int  rc = main1(argc, argv);
   return(rc);
} // main
