// $Id: pkcalls.c,v 1.2 1997/03/25 10:18:43 yotam Exp $
// TfmPk - TeXpk public calls. No X-Window calls.

#include <stdlib.h>
#include <strstream.h>
#include <tcl.h>
#include "texpk.h"
#include "block.h"
#include "tcltool.h"
#include "debug.h"


////////////////////////////////////////////////////////////////////////
bool
TfmpkProgBlock::openPk(const char * const *argv)
{
   bool rc = false;

   if (pPK == 0)
   {
      pPK = new TeXpk;
   }
   if (pPK != 0)
   {
      pPK->open(argv[0], atol(argv[1]));
      rc = pPK->active();
      if (rc)
      {
         tableDims();
         if ((_pickedChar < charFirst()) || (charLast() < _pickedChar) ||
             !pPK->hasChar(_pickedChar))
         {
            _pickedChar = charFirst();
         }
      }
      
   }
   _bitmapFocusDirty = true;
   _kernBitmapDirty = true;
   _dirtyBitmap = true;
   return(rc);
} // openPk


#if 0
////////////////////////////////////////////////////////////////////////
static unsigned int 
ratioMult(unsigned int n, unsigned int d, unsigned int x)
{
   unsigned int rc = (n * x + d - 1) / d;
   return(rc);
} // ratioMult
#endif


////////////////////////////////////////////////////////////////////////
void
TfmpkProgBlock::tableDims()
{
   if (!( (pPK != 0) && pPK->active() ))
   {
      return;
   }

   US16   minRow = charFirst() / 16, 
          maxRow = charLast() / 16;   

   US8  nRows = 0;
   if (minRow > maxRow)
   {
      _winWidth = _winHeight = 0;
   }
   else
   {
      unsigned int  cdp; // char draw pixels
      cdp =  pixelTableMult(maxRef(Left) + maxRef(Right));
      _winWidth = 16 * (cdp + chfill) + thfill;
      nRows = maxRow - minRow + 1;
      cdp =  pixelTableMult(maxRef(Up) + maxRef(Down));
      _winHeight = nRows * (cdp + cvfill) + tvfill;
   }
} // TfmpkProgBlock::tableDims


////////////////////////////////////////////////////////////////////////
bool
TfmpkProgBlock::pkCharInfo(int /* argc */, const char * const * const argv)
{
   bool  ok = (pPK != 0);
   if (ok)
   {
      US8  c = (US8)atoi(argv[0]);
      ok = pPK->hasChar(c);
      if (ok)           
      {
         ostrstream  ostr(interp->result, TCL_RESULT_SIZE - 1);
         ostr << pPK->dx(c) << ' ' << pPK->dy(c) << ' ' << 
                 pPK->bitmapWidth(c) << ' ' << pPK->bitmapHeight(c) << ' ' <<
                 pPK->hoff(c) << ' ' << pPK->voff(c) << ends;
      }
   }
   return(ok);
} // TfmpkProgBlock::pkCharInfo



////////////////////////////////////////////////////////////////////////
bool
TfmpkProgBlock::pkGeneral()
{
   ostrstream   result;
   bool  ok = (( (pPK != 0) && pPK->active() ));
   if (ok)
   {
      US32   ds = pPK->designSize();
      US32   hppp = pPK->hppp();
      US32   vppp = pPK->vppp();
      float  fds = ((float)ds / (float)(1<<20)); 
      float  fhppp = ((float)hppp / (float)(1<<16));
      float  fvppp = ((float)vppp / (float)(1<<16));
      result << 
         "Design-Size = " << fds << endl <<
         "hppp = " << fhppp << endl << 
         "vppp = " << fvppp << endl << 
         "comment = " << endl << "  " << pPK->comment() << endl << 
         "id = " << (int)pPK->id() << endl << 
         "Check-Sum = 0x" << hex << pPK->checkSum() << dec << ends;
   }
   setTclResult(interp, result);
   return(ok);
} // pkGeneral


////////////////////////////////////////////////////////////////////////
US8
TfmpkProgBlock::charFirst() const
{
   return(pPK ? pPK->charFirst() : 0);
} // TfmpkProgBlock::charFirst


////////////////////////////////////////////////////////////////////////
US8
TfmpkProgBlock::charLast() const
{
   return(pPK ? pPK->charLast() : 0);
} // TfmpkProgBlock::charLast


////////////////////////////////////////////////////////////////////////
US32
TfmpkProgBlock::maxRef(US8 lrud) const
{
   return(pPK ? pPK->maxOffset((TeXpk::Direction)lrud) : 0);
} // TfmpkProgBlock::maxRef



