// $Id: pkshow.c,v 1.6 1997/07/25 05:44:30 yotam Exp yotam $

#include <stdlib.h>
#include <math.h>
#include <X11/Xlib.h>

#include <tk.h>

#include <strstream.h>
#include <BitSet.h>

#include "FocusCharPixels.h"
#include "bignum.h"
#include "drawBitmap.h"
#include "texpk.h"
#include "textfm.h"
#include "block.h"
#include "tkinfo.h"
#include "xwin.h"
#include "debug.h"


class DWG
{
 public:
   DWG() {};
   DWG(Display* d, Window w, GC g) : dpy(d), win(w), gc(g) {};
   virtual ~DWG() {};

   Display* dpy;
   Window  win;
   GC      gc;
}; // DWG


// static const int  Pad = 3;


////////////////////////////////////////////////////////////////////////
static void
maximizeBy(US32& var, US32 val)
{
   if (var < val)
   {
      var = val;
   }
} // maximizeBy



////////////////////////////////////////////////////////////////////////
static void
drawCharPKrects
(
   Display* dpy,
   Window   win,
   GC       gc,
   TeXpk*   pPK,
   US8      c,
   int      xref,
   int      yref
)
{
   // assuming pixel to pixel

   // xref -= pPK->hoff(c);
   // yref -= pPK->voff(c);

   US32  ri, x, y, w, h;
   for (ri = pPK->nRects(c);  ri--;)
   {
      pPK->rect(x, y, w, h, c, ri);
      XFillRectangle(dpy, win, gc, xref + x, yref + y, w, h);      
   }
} // drawCharPKrects


////////////////////////////////////////////////////////////////////////
int
TfmpkProgBlock::pickChar(int argc, const char*  const * const argv)
{
   DASSERT(argc == 1 || argc == 2);
   DPRNL("pickChar");
   int  ok = 1;
   if (pPK && pPK->active())
   {
      int atoi0 = atoi(argv[0]);
      Int16  pickedChar = (Int16)(argc == 1 ? atoi0 : -1);
      if (argc == 2)
      {
         int  px = -1, py = -1;
         int  w = (int)winWidth();
         int  h = (int)winHeight();
         px = atoi0;
         py = atoi(argv[1]);
         ok = (0 <= px && px < w && 0 <= py && py < h);
         if (ok)
         {
            ok = 1;
            US8        nRows = (charLast()/16) - (charFirst()/16) + 1;
            int        cw = (w - thfill) / 16;
            int        ch = (h - tvfill) / nRows;
            US16 col = (px - thfill) / cw;   DASSERT(col < 16);
            US16 row = (py - tvfill) / ch;
            pickedChar = 16*(row + (charFirst() / 16)) + col;
         }
      }
      if ((charFirst() <= pickedChar) && (pickedChar <= charLast()) &&
          pPK->hasChar(pickedChar))
      {
         _pickedChar = pickedChar;
         DPRNL("display pick not yet implemented");
         _bitmapFocusDirty = true;
      }
   }

   return(ok);
} // pickChar


////////////////////////////////////////////////////////////////////////
// Generate the set of characters to be painted as ascending
static BitSet
ascendCharSet(const TeXtfm& tfm, US8 pickedChar)
{
   BitSet  s;
   s.clear();


   TeXtfm::CharInfo       charInfo = tfm.char_info(pickedChar);
   TeXtfm::CharInfo::Tag  tag = charInfo.tag();
   switch (tag)
   {
    case TeXtfm::CharInfo::list_tag:
      {
         Int16  charAscend = tfm.listAscend(pickedChar);
         s.set(charAscend);
      } 
      break;
    case TeXtfm::CharInfo::ext_tag:
      {
         US8  ei = charInfo.reminder();
         TeXtfm::ExtensibleRecipe  extRec = tfm.exten(ei);
         int  i;
         for (i = 4;  i--;)
         {
            US8  bi = extRec.byte(i);
            if ((bi != 0) || (i == TeXtfm::ExtensibleRecipe::rep))
            {
               s.set(bi);
            }
         }
      }
      break;
    default:;
   }
   return(s);
} // ascendCharSet


#define mxTfmPoint2PkPixel(v, m) \
        v = n1n2n3_div2p(hppp, ds, m * v, 16 + 20 + 20)

#define xTfmPoint2PkPixel(v) mxTfmPoint2PkPixel(v, 1)


#define myTfmPoint2PkPixel(v, m) \
        v = n1n2n3_div2p(vppp, ds, m * v, 16 + 20 + 20)

#define yTfmPoint2PkPixel(v) myTfmPoint2PkPixel(v, 1)


////////////////////////////////////////////////////////////////////////
void
TfmpkProgBlock::drawFontTable()
{
   if (pPK && pPK->active())
   {
      Tk_Window  tkwin = pTk->fontTableWin;
      Display*  dpy = Tk_Display(tkwin);
      Window   win = Tk_WindowId(tkwin);
      US16     width = Tk_Width(tkwin);
      US16     height = Tk_Height(tkwin);
      US16     reqWidth = Tk_ReqWidth(tkwin);
      US16     reqHeight = Tk_ReqHeight(tkwin);
      DPR("width=" << width << ", height=" << height <<  
          ", reqWidth=" << reqWidth << ", reqHeigt=" << reqHeight << endl);
      bool ok = (width == reqWidth) && (height == reqHeight);
      if (ok)
      {
         xwinfo->setBackingStore(tkwin);
         XFillRectangle(dpy, win, xwinfo->gc(XWinInfo::Clear), 
                        0, 0, width, height);            // clear
         GC         gcBorder = xwinfo->gc(XWinInfo::Border);
         GC         gcText = xwinfo->gc(XWinInfo::Text);
         int        w = (int)width;
         int        h = (int)height;
         US8        nRows = (charLast()/16) - (charFirst()/16) + 1;
         int        cw = (w - thfill) / 16;
         int        ch = (h - tvfill) / nRows;
         int        i, x, y;
         char xx[4];
         xx[1] = '0';
         for (i = 0, x = thfill;  i < 16;  i++, x += cw)
         {
            xx[0] = (i < 10 ? '0' : ('A' - 10)) + i;
            XDrawString(dpy, win, gcText, x + cw/2, 2*tvfill/3, xx, 1);
            XDrawLine(dpy, win, gcBorder, x, tvfill, x, h - 1);
         }
         (void)strcpy(xx, "xN0");
         for (i = 0, y = tvfill;  i < nRows;  i++, y += ch)
         {
            US8  xxn = (charFirst()/16) + i;
            xx[1] = (xxn < 10 ? '0' : ('A' - 10)) + xxn;
            XDrawString(dpy, win, gcText, 5, y + ch/2, xx, 3);
            XDrawLine(dpy, win, gcBorder, thfill, y, w - 1, y);
         }

         US16 c;
         US8  ri0 = charFirst() / 16;  
         GC   gcRaster = xwinfo->gc(XWinInfo::Raster);
         bool scale = (pixelFontPerScreenNominator != 
                       pixelFontPerScreenDenominator);
         unsigned  maxLeft = maxRef(Left);
         unsigned  maxUp   = maxRef(Up);
         if (scale)
         {
            updateBitmaps();
         }
         US32  xoff = thfill + chfill / 2;
         US32  yoff = tvfill + cvfill / 2;
         BitSet  ascendSet = ascendCharSet(*pTFM, _pickedChar);
         US16  cl = charLast();
         for (c = charFirst();  c <= cl;  c++)
         {
            GC   gcChar = gcRaster;
            if (c == _pickedChar)
            {
               gcChar = xwinfo->gc(XWinInfo::Select);
            } 
            else
            {
               if (ascendSet.test(c))
               {
                  gcChar = xwinfo->gc(XWinInfo::Ascend);
               }
            }
            if (pPK->hasChar(c))
            {
               US8  row = c / 16 - ri0,
                    col = c % 16;
               US32  xcoff = maxLeft - pPK->hoff(c);  // character
               US32  ycoff = maxUp   - pPK->voff(c);  // specific
               if (scale)
               {
                  xcoff = (xcoff * pixelFontPerScreenNominator) / 
                                   pixelFontPerScreenDenominator;
                  ycoff = (ycoff * pixelFontPerScreenNominator) / 
                                   pixelFontPerScreenDenominator;
                  drawBitmap(dpy, win, gcChar, &_bitmap[c],
                                 xoff + xcoff + col * cw,
                                 yoff + ycoff + row * ch);
               }
               else
               {
                  drawCharPKrects(dpy, win, gcChar, pPK, c, 
                                 xoff + xcoff + col * cw,
                                 yoff + ycoff + row * ch);
               }
            }
         }
      }
   }
} // TfmpkProgBlock::drawFontTable


////////////////////////////////////////////////////////////////////////
void
TkInfo::drawFontTable(Tk_Window DCODE(tkwin), ClientData appData)
{
   TfmpkProgBlock*  pb = (TfmpkProgBlock *)appData;
   DASSERT(tkwin == pb->pTk->fontTableWin);
   pb->drawFontTable();
} // TkInfo::drawFontTable


////////////////////////////////////////////////////////////////////////
void
XWinInfo::setBackingStore(Tk_Window  tkwin) const
{
   Display*  dpy = Tk_Display(tkwin);
   Screen*   screen = Tk_Screen(tkwin);   
   Window    win = Tk_WindowId(tkwin);
   if (XDoesBackingStore(screen))
   {
      XWindowAttributes  gwa;
      XGetWindowAttributes(dpy, win, &gwa);
      DPRNL("backing=" << gwa.backing_store << hex <<
            ", planes=" << gwa.backing_planes <<
            ", planes=" << gwa.backing_pixel << dec);

      XSetWindowAttributes  winAtt;
      winAtt.backing_store = WhenMapped;
      winAtt.backing_planes = 0;
      winAtt.backing_pixel = 0;
      for (int i = 0;  i < XWinInfo::Ngc;  i++)
      {
         GC  gci = gc(i);
         XGCValues   gcv;
         (void)XGetGCValues(dpy, gci, GCForeground,  &gcv);
         if (i == XWinInfo::Clear)
         {
            winAtt.backing_pixel = gcv.foreground;
         }
         else
         {
            winAtt.backing_planes |= gcv.foreground;
         }
      }

      unsigned long  valuemask = 0;
      if (winAtt.backing_store != gwa.backing_store)
      {
         valuemask |= CWBackingStore;
      }
      if (winAtt.backing_planes != gwa.backing_planes)
      {
         valuemask |= CWBackingPlanes;
      }
      if (winAtt.backing_pixel != gwa.backing_pixel)
      {
         valuemask |= CWBackingPixel;
      }
      if (valuemask)
      {
         XChangeWindowAttributes(dpy, win, valuemask, &winAtt);
      }
   }
} // setBackingStore


////////////////////////////////////////////////////////////////////////
int
TkInfo::configureFontTable(Tk_Window DCODE(tkwin), ClientData DCODE(appData))
{
   int             rc = TCL_OK;
#if DEBUG
   TfmpkProgBlock*  pb = (TfmpkProgBlock *)appData;
   DASSERT(tkwin == pb->pTk->fontTableWin);
#endif
   // Tk_GeometryRequest(tkwin, pb->winWidth() , pb->winHeight());
   return(rc);
} // configureFontTable


////////////////////////////////////////////////////////////////////////
void
TfmpkProgBlock::updateBitmaps()
{
   if (_dirtyBitmap)
   {
      _dirtyBitmap = false;
      if (pixelFontPerScreenNominator != pixelFontPerScreenDenominator)
      {
         US16  cf = charFirst();
         US16  cl = charLast();
         US16  c;
         for (c = cf;  c <= cl;  c++)
         {
            Bitmap  asisBitmap(pPK->bitmapWidth(c), pPK->bitmapHeight(c),
                               pPK->bitmapBitSetArray(c));
            _bitmap[c] = Bitmap(asisBitmap, pixelFontPerScreenNominator,
                                            pixelFontPerScreenDenominator);
#if DEBUG
{static bool dopr = false;
 if (dopr)  cout << "ASIS: " << endl <<
 asisBitmap << endl << "n/d:" << endl << _bitmap[c] << endl;
}
#endif
         }
      }
   }
} // updateBitmaps


////////////////////////////////////////////////////////////////////////
void
TfmpkProgBlock::focusBitmapUpdate()
{
   DPRNL("focusBitmapUpdate");
   if (_bitmapFocusDirty)
   {
      _bitmapFocusDirty = false;
      if (pPK != 0)
      {
         US8     c = _pickedChar;
         Bitmap  asisBitmap(pPK->bitmapWidth(c), pPK->bitmapHeight(c),
                            pPK->bitmapBitSetArray(c));
         _bitmapFocus = Bitmap(asisBitmap, pixelZoomScreenPerFont, 1);
      }
   }
} // focusBitmapUpdate


////////////////////////////////////////////////////////////////////////
bool
TfmpkProgBlock::fontLoaded() const
{
   bool  loaded = (pPK && pTFM) && pPK->active() && pTFM->active();
   return(loaded);
} // TfmpkProgBlock::fontLoadedfontLoaded


////////////////////////////////////////////////////////////////////////
bool
TfmpkProgBlock::setFocusCharPixels()
{
   bool      ok = false;
   US8       c = pickedChar();
   DPRNL("drawFocusChar=" << hex << (US16)c << dec);
   focusBitmapUpdate();
   if (fontLoaded())
   {
      ok = true;
      FocusCharPixels&  fcPixels = *xwinfo->focusCharPixels();
      US32     xpre;
      unsigned mult = pixelZoomScreenPerFont;

      {
         const TeXtfm&  tfm = *pTFM;  // most values are Fixed-Point 2^20

         // convert to tfm-points:     tp(x) = ds * x / 2^20
         // tfm-points to pk-pixels:   pp(t) = [hv]ppp * t / 2^16
         US32  ds = tfm.hDesignSize();
         US32  hppp = pPK->hppp();
         US32  vppp = pPK->vppp();

         US32  ExtSpc    = tfm.param(TeXtfm::extra_space_code);
         US32  SpcStrch  = tfm.param(TeXtfm::space_stretch_code);
         US32  WordSpace = tfm.param(TeXtfm::space_code);
         US32  SpcShrink = tfm.param(TeXtfm::space_shrink_code);
         US32  uslant    = tfm.param(TeXtfm::slant_code);
         float fslant    = (float)TeXtfm::FixWord(uslant);
         float aslant    = fabs(fslant);
         US32  slantUp   = (US32)(aslant * (float)maxRef(Up));
         US32  slantDown = (US32)(aslant * (float)maxRef(Down));

         xpre = ExtSpc + SpcStrch + WordSpace;
         mxTfmPoint2PkPixel(xpre, mult);

         Int32 icw = tfm.cWidth(c);
         Int32 ich = tfm.cHeight(c);
         Int32 icd = tfm.cDepth(c);
         US32  cw = (icw > 0 ? icw : 0);
         US32  ch = (ich > 0 ? ich : 0);
         US32  cd = (icd > 0 ? icd : 0);
         mxTfmPoint2PkPixel(cw, mult);
         myTfmPoint2PkPixel(ch, mult);
         myTfmPoint2PkPixel(cd, mult);

         fcPixels.bbl = _pad + xpre + mult * maxRef(Left);
         fcPixels.bbr = fcPixels.bbl + cw;
         fcPixels.bbm = _pad + mult * maxRef(Up);
         fcPixels.bbh = fcPixels.bbm - ch;
         fcPixels.bbd = fcPixels.bbm + cd;

         fcPixels.slantFlag = (uslant != 0);
         if (fcPixels.slantFlag)
         {
            fcPixels.bbstl = fcPixels.bbsdl = fcPixels.bbl;
            fcPixels.bbstr = fcPixels.bbsdr = fcPixels.bbr;
            bool  positive = (fslant > 0.);
            if (!positive)
            {
               slantUp = -slantUp;
               slantDown = -slantDown;
            }
            fcPixels.bbstl += slantUp;
            fcPixels.bbstr += slantUp;
            fcPixels.bbsdl -= slantDown;
            fcPixels.bbsdr -= slantDown;
         }

         // spacings 
         mxTfmPoint2PkPixel(SpcStrch, mult);
         mxTfmPoint2PkPixel(SpcShrink, mult);
         mxTfmPoint2PkPixel(WordSpace, mult);
         fcPixels.dy = (SpcStrch + SpcShrink + _pad) / 2;

         // spacings to the left.
         fcPixels.leftSpace = fcPixels.bbl - WordSpace;
         fcPixels.leftStretch = fcPixels.leftSpace - SpcStrch;
         fcPixels.leftShrink  = fcPixels.leftSpace + SpcShrink;

         if (fcPixels.extSpcFlag = (ExtSpc != 0))
         {
            mxTfmPoint2PkPixel(ExtSpc, mult);
            unsigned  rh = ExtSpc + _pad;
            // fcPixels.extSpcRight = spcPos - SpcStrch;
            fcPixels.extSpcLeft  = fcPixels.leftStretch - ExtSpc;
            fcPixels.extSpcTop   = fcPixels.bbm - rh;
         }

         // spacings to the right.

         US32  cital = tfm.cItalic(c);
         mxTfmPoint2PkPixel(cital, mult);
         fcPixels.itCor = fcPixels.bbr + cital;

         fcPixels.rightSpace = fcPixels.bbr + WordSpace;
         fcPixels.rightShrink  = fcPixels.rightSpace - SpcShrink;
         fcPixels.rightStretch = fcPixels.rightSpace + SpcStrch;

         // ex
         unsigned&  ex = fcPixels.ex;
         ex = tfm.param(TeXtfm::x_height_code);
         myTfmPoint2PkPixel(ex, mult);
         fcPixels.exTop = fcPixels.bbm - ex;

         // em
         unsigned&  em = fcPixels.em;
         em = tfm.param(TeXtfm::quad_code);
         mxTfmPoint2PkPixel(em, mult);
         fcPixels.emRight = fcPixels.bbl + em;

         // raster offset
         Int32     hoff = pPK->hoff(c);
         Int32     voff = pPK->voff(c);
         fcPixels.xoff = _pad + xpre + mult * (maxRef(Left) - hoff);
         fcPixels.yoff = _pad +        mult * (maxRef(Up)   - voff);
      }

   }
   return(ok);
} // TfmpkProgBlock::setFocusCharPixels


////////////////////////////////////////////////////////////////////////
bool
TfmpkProgBlock::focusCharWinSize()
{
   bool  ok = false;
   focusBitmapUpdate();
   if (pPK && pPK->active() && pTFM)
   {
      ok = true;
      ostrstream  ostr(interp->result, TCL_RESULT_SIZE - 1);
      // ostr << _bitmapFocus.w() << ' ' <<  _bitmapFocus.h() << ends;
      tableDims();

      // The maximal BBox for the bitmap
      US32  w  = maxRef(Left) + maxRef(Right);
      US32  h  = maxRef(Up) + maxRef(Down);

      // See Drawing. Horizontally we have:
  //
  //                                     [Slant] 
  //  [ExtSpc][SpcStrch][WordSpace][BBoxX][WordSpace][SpcStrch] 
  //                    |    [Slant]     [Itlic]    |
  //                    [SpcShrink]       [SpcShrink]
  //
  //  We should secure space
           
      const TeXtfm&   tfm = *pTFM;  // most values are Fixed-Point 2^20

      US32  ExtSpc    = tfm.param(TeXtfm::extra_space_code);
      US32  SpcStrch  = tfm.param(TeXtfm::space_stretch_code);
      US32  WordSpace = tfm.param(TeXtfm::space_code);

      US32  uslant    = tfm.param(TeXtfm::slant_code);
      float fslant    = (float)TeXtfm::FixWord(uslant);
      float aslant    = fabs(fslant);
      US32  slantUp   = (US32)(aslant * (float)maxRef(Up));
      US32  slantDown = (US32)(aslant * (float)maxRef(Down));
      US32  slantMax  = (slantUp < slantDown ? slantDown : slantUp);

      US32  xpre = ExtSpc + SpcStrch + WordSpace;
      maximizeBy(xpre, slantMax);

      US32 xpost = WordSpace + SpcStrch;
      maximizeBy(xpost, slantMax);

      // Ensuring italic correction -This is probably will never be needed.
      int   i;  
      US32  itCorMax = 0;
      for (i = tfm.length(TeXtfm::ni);  i--;)
      {
         US32  itCor = tfm.italic(i);
         maximizeBy(itCorMax, itCor);
      }
      maximizeBy(xpost, itCorMax);

      // convert to tfm-points:     tp(x) = ds * x / 2^20
      // tfm-points to pk-pixels:   pp(t) = [hv]ppp * t / 2^16
      US32  ds = tfm.hDesignSize();
      US32  hppp = pPK->hppp();
      mxTfmPoint2PkPixel(xpre, pixelZoomScreenPerFont);
      mxTfmPoint2PkPixel(xpost, pixelZoomScreenPerFont);

      US32  wMax = tfm.maxWidth();
      US32  hMax = tfm.maxHeight();
      mxTfmPoint2PkPixel(wMax, pixelZoomScreenPerFont);
      mxTfmPoint2PkPixel(hMax, pixelZoomScreenPerFont);
      maximizeBy(w, wMax);
      maximizeBy(h, hMax);

      w *= pixelZoomScreenPerFont;
      h *= pixelZoomScreenPerFont;
      w += (xpre + xpost);
      
      w += 2*_pad; // Pad pixels padx
      h += 2*_pad; // Pad pixels pady

      ostr << w << ' ' << h << ends;
   }
   return(ok);
} // TfmpkProgBlock::focusCharWinSize


////////////////////////////////////////////////////////////////////////
bool
TfmpkProgBlock::focusCharProbe(const char* const * const argv)
{
   bool  ok = true;
   ostrstream  result(interp->result, TCL_RESULT_SIZE - 1);
   int  x = atoi(argv[0]);
   int  y = atoi(argv[1]);
   Tk_Window  tkwin = pTk->winFocusChar;
   int  w = Tk_Width(tkwin);
   int  h = Tk_Height(tkwin);
   result << xwinfo->focusCharPixels()->probeMessage(x, y, w, h, _pad) << ends;
   return(ok);
} // TfmpkProgBlock::focusCharProbe


////////////////////////////////////////////////////////////////////////
static void
diamondCut(Display* dpy, Window win, GC gc, const XPoint* xpt4)
{
   int  i;
   for (i = 4;  i--;)
   {
      int  i1 = (i + 1) % 4;
      XDrawLine(dpy, win, gc, xpt4[i].x, xpt4[i].y, xpt4[i1].x, xpt4[i1].y);
   } 
   XDrawLine(dpy, win, gc, xpt4[1].x, xpt4[1].y, xpt4[3].x, xpt4[3].y);
} // diamondCut


////////////////////////////////////////////////////////////////////////
void
TfmpkProgBlock::drawFocusChar()
{
   DCODE(US8       c = pickedChar());
   DPRNL("drawFocusChar=" << hex << (US16)c << dec);
   focusBitmapUpdate();
   if (fontLoaded())
   {
      (void)setFocusCharPixels(); // we may be able to save this off redraw
      const FocusCharPixels&  fcp = *xwinfo->focusCharPixels();
      Tk_Window  tkwin = pTk->winFocusChar;
      // Tk_GeometryRequest(tkwin, _bitmapFocus.w(), _bitmapFocus.h());
      Display*  dpy = Tk_Display(tkwin);
      Window   win = Tk_WindowId(tkwin);

      {
         GC       gcClear = xwinfo->gc(XWinInfo::Clear);
         XFillRectangle(dpy, win, gcClear, 
                        0, 0, Tk_Width(tkwin), Tk_Height(tkwin));
      }

      if (_focusGrid)
      {
         GC  gcBrd = xwinfo->gc(XWinInfo::Border);

         // horizontal lines
         XDrawLine(dpy, win, gcBrd, fcp.bbl, fcp.bbh, fcp.bbr, fcp.bbh); // ---
         XDrawLine(dpy, win, gcBrd, fcp.bbl, fcp.bbm, fcp.bbr, fcp.bbm); // -_-
         XDrawLine(dpy, win, gcBrd, fcp.bbl, fcp.bbd, fcp.bbr, fcp.bbd); // ___

         // vertical lines
         XDrawLine(dpy, win, gcBrd, fcp.bbl, fcp.bbh, fcp.bbl, fcp.bbd); // |
         XDrawLine(dpy, win, gcBrd, fcp.bbr, fcp.bbh, fcp.bbr, fcp.bbd); //    |

         if (fcp.slantFlag)
         {
            // top       /-/
            XDrawLine(dpy, win, gcBrd, 
                      fcp.bbl,  fcp.bbm, fcp.bbstl, fcp.bbh);   //  /
            XDrawLine(dpy, win, gcBrd, 
                      fcp.bbstl, fcp.bbh, fcp.bbstr, fcp.bbh);  //   -
            XDrawLine(dpy, win, gcBrd, 
                      fcp.bbstr, fcp.bbh, fcp.bbr,  fcp.bbm);   //    /

            // bottom  /_/
            XDrawLine(dpy, win, gcBrd, 
                      fcp.bbl,  fcp.bbm, fcp.bbsdl, fcp.bbd);   //  /
            XDrawLine(dpy, win, gcBrd, 
                      fcp.bbsdl, fcp.bbd, fcp.bbsdr, fcp.bbd);  //   _
            XDrawLine(dpy, win, gcBrd, 
                      fcp.bbsdr, fcp.bbd, fcp.bbr,  fcp.bbm);   //    /
         }

         // spacings 
         XPoint  xpt[4];
         
         // spacings to the left.
         XDrawLine(dpy, win, gcBrd, fcp.bbl, fcp.bbm, fcp.leftShrink, fcp.bbm);

          xpt[0].x = fcp.leftShrink;  xpt[0].y = fcp.bbm; 
          xpt[1].x = fcp.leftSpace;   xpt[1].y = fcp.bbm - fcp.dy;
          xpt[2].x = fcp.leftStretch; xpt[2].y = fcp.bbm;
          xpt[3].x = fcp.leftSpace;   xpt[3].y = fcp.bbm + fcp.dy;
         diamondCut(dpy, win, gcBrd, xpt);
         
         if (fcp.extSpcFlag)
         {
            XDrawRectangle(dpy, win, gcBrd, 
                           fcp.extSpcLeft, fcp.extSpcTop,
                           (fcp.leftStretch - fcp.extSpcLeft),
                           (fcp.bbm - fcp.extSpcTop));
         }

         // spacings to the right.      // xpt[].y as was
         XDrawLine(dpy, win, gcBrd, fcp.bbr, fcp.bbm, fcp.rightShrink, fcp.bbm);
          xpt[0].x = fcp.rightStretch;  // xpt[0].y = fcp.bbm; 
          xpt[1].x = fcp.rightSpace;    // xpt[1].y = fcp.bbm - fcp.dy;
          xpt[2].x = fcp.rightShrink;   // xpt[2].y = fcp.bbm;
          xpt[3].x = fcp.rightSpace;    // xpt[3].y = fcp.bbm + fcp.dy;
         diamondCut(dpy, win, gcBrd, xpt);

         // italic correctopn
         XDrawLine(dpy, win, gcBrd, 
                   fcp.itCor, fcp.bbm, fcp.itCor, fcp.bbm - fcp.dy);

         // ex
         XFillRectangle(dpy, win, gcBrd, 0, fcp.exTop, _pad, fcp.ex);
         XFillRectangle(dpy, win, gcBrd, 
                        Tk_Width(tkwin) - _pad - 1, fcp.exTop, _pad, fcp.ex);
         
         // em
         XFillRectangle(dpy, win, gcBrd, fcp.bbl, 0,      fcp.em, _pad);
         XFillRectangle(dpy, win, gcBrd, 
                        fcp.bbl, Tk_Height(tkwin) - _pad - 1, fcp.em, _pad);
         
      }

      {
         GC        gcRaster = xwinfo->gc(XWinInfo::Raster);
         drawBitmap(dpy, win, gcRaster, &_bitmapFocus, fcp.xoff, fcp.yoff);
      }
   }
} // TfmpkProgBlock::drawFocusChar


////////////////////////////////////////////////////////////////////////
int
TfmpkProgBlock::configureFocusChar()
{
   int             rc = TCL_OK;
   DPRNL("configureFocusChar");
   // Tk_GeometryRequest(pTk->winFocusChar, 
   //                    _bitmapFocus.w(), _bitmapFocus.h());
   return(rc);
} // TfmpkProgBlock::configureFocusChar


////////////////////////////////////////////////////////////////////////
void
TkInfo::drawFocusChar(Tk_Window, ClientData appData)
{
   TfmpkProgBlock*  pb = (TfmpkProgBlock *)appData;
   DPRNL("drawFocusChar");
   pb->drawFocusChar();
} // TkInfo::drawFocusChar


////////////////////////////////////////////////////////////////////////
int
TkInfo::configureFocusChar(Tk_Window, ClientData appData)
{
   int             rc = TCL_OK;
   TfmpkProgBlock*  pb = (TfmpkProgBlock *)appData;
   DPRNL("configureFocusChar");
   rc = pb->configureFocusChar();
   return(rc);
} // configureFocusChar


