// $Id: pushdir.c,v 1.1 1995/10/01 22:22:04 yotam Exp $
#include <limits.h>
#include <unistd.h>
#include <sys/types.h>

#include <iostream.h>

#include "pushdir.h"
#include "debug.h"


////////////////////////////////////////////////////////////////////////
const char *  PushDir::trivial = ""; 


////////////////////////////////////////////////////////////////////////
PushDir::PushDir(const char *dirname):
   savedDir(0),
   _dir(0)
{
   DPR("PushDir(" << dirname << ")" << endl);
   if (*dirname == '\0')
   {
      savedDir = (char *)trivial;
   }
   else
   {
      size_t  nextBufSize, bufSize;

      savedDir = getcwd(autoBuf, sizeof(autoBuf) - 1);
      int  gotwd = (savedDir != 0);
      nextBufSize = 2 * sizeof(autoBuf);
      while (!gotwd && (nextBufSize < (ULONG_MAX / 2)))
      {
         bufSize = nextBufSize;
         savedDir = new char[bufSize];
         gotwd = ((savedDir != 0) && (getcwd(savedDir, bufSize-1) != 0));
         if (!gotwd)
         {
            nextBufSize *= 2;
            if (savedDir)
            {
               delete [] savedDir;  savedDir = 0;
            }
         }
      }

      int  ok = gotwd && 
                (chdir(dirname) == 0) &&
                ((_dir = opendir(".")) != 0);
      if (!ok && (savedDir != 0))
      {
         if (savedDir != autoBuf)
         {
            delete [] savedDir;
         }
         savedDir = 0;
      }
   }
} // PushDir::PushDir


////////////////////////////////////////////////////////////////////////
PushDir::~PushDir()
{
   DPR("~PushDir, savedDir=" << savedDir << endl);
   pop();
} // PushDir::~PushDir


////////////////////////////////////////////////////////////////////////
void
PushDir::pop()
{
   DPR("PushDir::pop, savedDir=" << savedDir << endl);
   if (_dir)
   {
      closedir(_dir);  _dir = 0;
   }
   if ((savedDir != 0) && (savedDir != trivial)) 
   {
      if (chdir(savedDir) != 0)
      {
         cerr << "Failed to chdir(" << savedDir << ")" << endl; 
      }
      if (savedDir != autoBuf)
      {
         delete [] savedDir;
      }
      savedDir = 0;
   }
} // PushDir::~PushDir
