// $Id: tfminfo.c,v 1.4 1997/07/11 05:06:00 yotam Exp $

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <strstream.h>
#include <iomanip.h>
#include <tcl.h>
#include "block.h"
#include "tcltool.h"
#include "textfm.h"
#include "debug.h"


////////////////////////////////////////////////////////////////////////
static void
showFixedWord
(
   ostream&                     result,
   int                         argc,
   const char*  const*  const  argv,
   US32                        w32
)
{
   DPRNL("prec=" << result.precision());
   if ((argc >= 3) && !(strcmp(argv[2], "float")))
   {
      result << (float)TeXtfm::FixWord(w32) << ends;
   }
   else
   {
      result << w32 << ends;
   }
} // showFixedWord


////////////////////////////////////////////////////////////////////////
// for the ostream& << only
class H2
{
 public:
   H2(US8 c, const char* zeroDisp = 0): _c(c), _zd(zeroDisp) {};
   US8  _c;
   const char* _zd;
}; // H2

////////////////////////////////////////////////////////////////////////
static ostream&
operator << (ostream& os, H2 h2)
{
   DASSERT(os.flags() & ios::hex);
   if ((h2._c == 0) && h2._zd)
   {
      os << h2._zd;
   }
   else
   {
      os << 'x' << setw(2) << (unsigned)h2._c;
   }
   return(os);
} // ostream << H2



////////////////////////////////////////////////////////////////////////
static void
general(ostream& os, TeXtfm& tfm)
{
#define FltParam(pcode)  ((float) (TeXtfm::FixWord(tfm.param(TeXtfm::pcode))) )
   os << "Design-Size = " << tfm.designSize() << endl << 
         "Slant = " << FltParam(slant_code) << endl <<
         "Word-Space = " << FltParam(space_code) << endl <<
         "Space-Strech = " << FltParam(space_stretch_code) << endl <<
         "Space-Shrink = " << FltParam(space_shrink_code) << endl <<
         "ex-Height = " << FltParam(x_height_code) << endl <<
         "quad=em = " << FltParam(quad_code) << endl <<
         "+ Sentence-Space = " << FltParam(extra_space_code) << endl;

   unsigned lh = tfm.length(TeXtfm::lh);
   if (lh > 2)
   {
      unsigned  nc = lh * sizeof(US32);
      char*      extra = new char[nc + 1];
      (void)memcpy(extra, tfm.extraHeader(), nc);
      int  i;
      for (i = nc;  i--;)
      {
         unsigned char  c = extra[i];
         if (c && !isprint(c))
         {
            extra[i] = ' ';
         }
      }
      extra[nc] = '\0';
      os << "HeaderExtra = '" << extra << "'," << endl;
      delete [] extra;
   }
   os << "Check-Sum = 0x" << hex << tfm.header(0) << dec << ends;

#if 0
   os << "Check-Sum=0x" << hex << tfm.header(0) << dec << 
         ",  Design-Size=" << (float)(TeXtfm::FixWord(tfm.header(1))) << 
         "," << endl;
   unsigned lh = tfm.length(TeXtfm::lh);
   if (lh > 2)
   {
      unsigned  nc = lh * sizeof(US32);
      char*      extra = new char[nc + 1];
      (void)memcpy(extra, tfm.extraHeader(), nc);
      int  i;
      for (i = nc;  i--;)
      {
         char  c = extra[i];
         if (c && !isprint(c))
         {
            extra[i] = ' ';
         }
      }
      extra[nc] = '\0';
      os << "HeaderExtra='" << extra << "'," << endl;
      delete extra;
   }

#define FltParam(pcode)  ((float) (TeXtfm::FixWord(tfm.param(TeXtfm::pcode))) )
   os << 
         "Slant=" << FltParam(slant_code) <<
      ",  Word-Space=" << FltParam(space_code) <<
      ",  Space-Strech=" << FltParam(space_stretch_code) <<
      ",  Space-Shrink=" << FltParam(space_shrink_code) <<
      ",  ex-Height=" << FltParam(x_height_code) <<
      ",  quad=em=" << FltParam(quad_code) <<
      ",  Extra-Sentence-Space=" << FltParam(extra_space_code);
#endif
} // general


////////////////////////////////////////////////////////////////////////
bool
TfmpkProgBlock::tfmInfo(int argc, const char*  const*  const argv)
{
   bool         ok = true;
   int          index = -1; // inited to silent warning
   ostrstream   result;
   TeXtfm*       ptfm = pTFM;
   int          oldPrec = result.precision(4);
   if (argc == 0)
   {
      ok = false;
      cerr << "tfminfo missing option" << endl;
   }
   else if (ptfm == 0)
   {
      ok = false;
      cerr << "TFM not loaded";
   }

   // General Info
   else if (!strcmp(argv[0], "general"))
   {
      general(result, *ptfm);  result << ends;
   }
   
   
   // char info
   else if (!strcmp(argv[0], "charinfo") && (argc >= 2) &&
            (ptfm->length(TeXtfm::bc) <= (index = atoi(argv[1]))) &&
            (index  <= ptfm->length(TeXtfm::ec))
           )
   {
      TeXtfm::CharInfo  cinf = ptfm->char_info(index);
      result << (int)cinf.width_index()  << ' ' <<
                (int)cinf.height_index() << ' ' <<
                (int)cinf.depth_index()  << ' ' <<
                (int)cinf.italic_index() << ' ' <<
                (int)cinf.tag()          << ' ' <<
                (int)cinf.reminder()     << ends;
   }


   // Width
   else if (!strcmp(argv[0], "nwidths"))
   {
      result << ptfm->length(TeXtfm::nw) << ends;
   }
   else if (!strcmp(argv[0], "width") && (argc >= 2) &&
            (0 <= (index = atoi(argv[1]))) &&
            (index  < ptfm->length(TeXtfm::nw))
           )
   {
      showFixedWord(result, argc, argv, ptfm->width(index) );
   }

   // Height
   else if (!strcmp(argv[0], "nheights"))
   {
      result << ptfm->length(TeXtfm::nh) << ends;
   }
   else if (!strcmp(argv[0], "height") && (argc >= 2) &&
            (0 <= (index = atoi(argv[1]))) &&
            (index  < ptfm->length(TeXtfm::nh))
           )
   {
      showFixedWord(result, argc, argv, ptfm->height(index) );
   }

   // Depth
   else if (!strcmp(argv[0], "ndepths"))
   {
      result << ptfm->length(TeXtfm::nd) << ends;
   }
   else if (!strcmp(argv[0], "depth") && (argc >= 2) &&
            (0 <= (index = atoi(argv[1]))) &&
            (index  < ptfm->length(TeXtfm::nd))
           )
   {
      showFixedWord(result, argc, argv, ptfm->depth(index) );
   }

   // Italic
   else if (!strcmp(argv[0], "nitalics"))
   {
      result << ptfm->length(TeXtfm::ni) << ends;
   }
   else if (!strcmp(argv[0], "italic") && (argc >= 2) &&
            (0 <= (index = atoi(argv[1]))) &&
            (index  < ptfm->length(TeXtfm::ni))
           )
   {
      showFixedWord(result, argc, argv, ptfm->italic(index) );
   }

   // Lig-Kern
   else if (!strcmp(argv[0], "nligkerns"))
   {
      result << ptfm->length(TeXtfm::nl) << ends;
   }
   else if (!strcmp(argv[0], "ligkern") && (argc >= 2) &&
            (0 <= (index = atoi(argv[1]))) &&
            (index  < ptfm->length(TeXtfm::nl))
           )
   {
      char  oldFill = result.fill('0');
      TeXtfm::LigKern  ligKern = ptfm->lig_kern(index);
      // owner(s)
      US8 cown;
      unsigned nOwners = ptfm->ligKernOwners(&cown, 1, index);
      DASSERT(nOwners >= 1);
      US8              cnext   = ligKern.next();
      result << hex;
      if (nOwners > 1)
      {
         result << "x**";
      }
      else if (nOwners == 1)
      {
         result << H2(cown);
      } 
      else // nOwners == 0
      {
         result << "---";
      }
      result << ": " << H2(cnext) << " ";
      result << (((nOwners == 1) && isprint(cown)) ? (char)cown : ' ');
      result << (isprint(cnext) ? (char)cnext : ' ');
      result << '>';
      US8  rem = ligKern.reminder();
      if (ligKern.kern_flag())
      {
         ok = (rem < ptfm->length(TeXtfm::nk));
         if (ok)
         {
            result << " K " << dec << (int)rem << 
                   // ' ' << (float)TeXtfm::FixWord( ptfm->kern( rem ) ) << 
                   ends;
         }
      }
      else
      {
         ok = (ptfm->length(TeXtfm::bc) <= rem &&
               rem <= ptfm->length(TeXtfm::ec));
         result << " L " << H2(rem) << dec << ends;
      }
      (void)result.fill(oldFill);
   }

   else if (!strcmp(argv[0], "ligkernOwners") && (argc >= 2) &&
            (0 <= (index = atoi(argv[1]))) &&
            (index  < ptfm->length(TeXtfm::nl))
           )
   {
      const unsigned  maxOwn = TCL_RESULT_SIZE / 4;
      US8             owners[maxOwn]; 
      unsigned        nOwners = ptfm->ligKernOwners(owners, maxOwn, index);
      if (nOwners > maxOwn)
      {
         nOwners = maxOwn;
      }
      unsigned  i;
      for (i = 0;  i < nOwners;  i++)
      {
         result << ' ' << (US16)owners[i];
      }
      result << ends;
   }

   // Kern
   else if (!strcmp(argv[0], "nkerns"))
   {
      result << ptfm->length(TeXtfm::nk) << ends;
   }
   else if (!strcmp(argv[0], "kern") && (argc >= 2) &&
            (0 <= (index = atoi(argv[1]))) &&
            (index  < ptfm->length(TeXtfm::nk))
           )
   {
      showFixedWord(result, argc, argv, ptfm->kern(index) );
   }

   // Extensible Recipes
   else if (!strcmp(argv[0], "nextens"))
   {
      result << ptfm->length(TeXtfm::ne) << ends;
   }
   else if (!strcmp(argv[0], "exten") && (argc >= 2) &&
            (0 <= (index = atoi(argv[1]))) &&
            (index  < ptfm->length(TeXtfm::ne))
           )
   {
      TeXtfm::ExtensibleRecipe  er = ptfm->exten(index);
      unsigned  top = er.byte(er.top);
      unsigned  mid = er.byte(er.mid);
      unsigned  bot = er.byte(er.bot);
      unsigned  rep = er.byte(er.rep);
      char  oldFill = result.fill('0');
      result << hex << 
         H2(top, "---") << ' ' << 
         H2(mid, "---") << ' ' << 
         H2(bot, "---") << ' ' << 
         H2(rep, "---");
      (void)result.fill(oldFill);
      result << dec << ends;
   }

   // Param
   else if (!strcmp(argv[0], "nparams"))
   {
      result << ptfm->length(TeXtfm::np) << ends;
   }
   else if (!strcmp(argv[0], "param") && (argc >= 2) &&
            (0 <= (index = atoi(argv[1]))) &&
            (index  < ptfm->length(TeXtfm::np))
           )
   {
      showFixedWord(result, argc, argv, ptfm->param(index) );
   }

   // file-size
   else if (!strcmp(argv[0], "sizeOK"))
   {
      result << (int)ptfm->sizeOk() << ends;
   }


   else
   {
      result << "tfminfo - bad option '" << argv[0] << "'" << endl;
      ok = false;
   }
   (void)result.precision(oldPrec);
   setTclResult(interp, result);

   return(ok);
} // tfmInfo
