# $Id: tfmpk.tcl,v 1.11 1997/07/25 05:44:30 yotam Exp yotam $

#############################
# TeX's tfm+pk files viewer #
#############################

set debug 0;
set DEBUG 0;
if {$DEBUG} {

 ########################################################################
 proc dputs {args} {
    # puts stderr $args;
    puts  $args;
 }; # dputs


 ########################################################################
 proc show {current} {
   if {$current > 0} {
     set info [info level $current]
     set proc [lindex $info 0]
     puts stderr "$current: Procedure $proc \
                  {[info args $proc]}"
     set index 0
     foreach arg [info args $proc] {
       puts stderr \
            "\t$arg = [lindex $info [incr index]]"
     }
   } else {
     puts stderr "Top level"
   }
 }; # show


 ########################################################################
 proc breakpoint {} {
   set max [expr [info level] - 2];
   set current $max;
   show $current;
     while {1} {
     puts -nonewline stderr "#$current: "
     gets stdin line
     while {![info complete $line]} {
       puts -nonewline stderr "? "
       append line \n[gets stin]
     }
     switch -- $line {
       + {if {$current < $max} {
           show [incr current]
           }
         }
       - {if {$current < $max} {
           show [incr current -1]
           }
         }
       C {puts stderr "Resuming execution"; return}
       ? {show $current}
       default {
         catch { uplevel $current $line } result
         puts stderr $result
       }
     }
   }
 }; # breakpoint


 ########################################################################
 proc dassert {cond {msg "Assertion Failed"}} {
    if {!$cond} {
       dputs stderr $msg;
       breakpoint;
    }
 }; # dassert

} else { # ! DEBUG - dummies
 proc dputs {args} {};
 proc dassert {cond {msg ""}} {};
}


########################################################################
# maximize variable by value, also return maxed-value
########################################################################
proc maximize {varName val} {
   upvar $varName var;
   if {$var < $val} {
      set var $val;
   }
   return $var;
}; # maximize


########################################################################
# like x ? y : z - but y,z are any string
proc sifthenelse {cond sTrue sFalse} {
   set s $sFalse;
   if {$cond} {
      set s $sTrue;
   }
   return $s;
}; # sifthenelse


####################
proc basename {fn} {
   dputs "basename fn=$fn";
   set bn $fn;
   set slash [string last / $fn];
   if {$slash >= 0} {
      set bn [string range $fn [expr $slash+1] end];
   }
   return $bn;
}; # basename


########################################################################
# Capitalize first letter
proc capitalFirst {word} {
   set fc [string range $word 0 0];
   set tail [string range $word 1 end];
   append cap [string toupper $fc] $tail;
   return $cap;
}; # capitalFirst


########################################################################
# If exist then destroy. Also save the window location.
########################################################################
proc safeDestroy {w} {
   global tfmpkGlob;
   if {[winfo exists $w]} {
      #dpts "destroy $w";
      set g [wm geometry $w];
      set gLoc [string range $g [string first + $g]  end];
      dputs "gLoc=$gLoc";
      set tfmpkGlob(geoloc,$w) $gLoc;
      destroy $w;
   }
}; # safeDestroy


########################################################################
# Given basename - find file
proc findFile {base} {
   global  tfmpkGlob env;
   set     result "";
   if [file exists $base] {  # absolute or in local directory
      set result $base;
   } else {

      # Add executable directory to the search path
      set slash [string last / $tfmpkGlob(arg0)];
      set searchPath $env(PATH);
      if {$slash >= 0} {
         set dir [string range $tfmpkGlob(arg0) 0 [expr $slash-1]];
         set searchPath [format "%s:%s" $dir $searchPath];
      }
      dputs "arg0=$tfmpkGlob(arg0), searchPath=$searchPath";

      while {[string compare $searchPath ""] && ![string compare $result ""]} {
         set colon [string first : $searchPath];
         if {$colon >= 0} {
            set dir [string range $searchPath 0 [expr $colon - 1]];
            set searchPath [string range $searchPath [expr $colon + 1] end];
         } else {
            set dir $searchPath;
            set searchPath "";
         }
         set candid [format %s/%s $dir $base];
         if {[file exists $candid]} {
            set result $candid;
         }
      }
   }
   return $result;
}; # findFile


######################
# environment override
proc env_override {varname envname} {
   global debug env
   set envexist 0
   set env_names [array names env]
   set ei [llength $env_names]; # number of environment variables
   while {$ei && ($envexist == 0)} {
      set ei [expr $ei - 1]
      set envi [lindex $env_names $ei]
      set envexist [expr ![string compare  $envname $envi]]
   }
   # set envexist [expr info exist env($envname)]
   if ($debug) {
      if ($envexist) {
         puts "override $varname by env($envname) $env($envname)"
      } else {
         puts "$varname not overrided by env($envname)"
      }
   }
   if ($envexist) {
      upvar $varname  vn
      set vn $env($envname)
   }
}; # env_override


###################
# Initialze Globals
proc initGlobals {} {
   global tfmpkGlob;

   dputs "initGlobals";
   # Is there data loaded?
   set tfmpkGlob(dataLoaded) 0;

   set tfmpkGlob(fontList) {};

   set tfmpkGlob(color,TFM) tan;
   set tfmpkGlob(color,PK)  pink2;

   set tfmpkGlob(namePattern) "";
   set tfmpkGlob(dpiPattern) "";

   set tfmpkGlob(serial) 0;

   # Focus character
   set tfmpkGlob(focusGridOn) 1;

   # Use bitmap Icon? (instead of color-window-icon)
   set tfmpkGlob(bitmapIcon) 0;
               
   # application flags
    set tfmpkGlob(flags) 0;

   # TFM - info

   set tfmpkGlob(tfmParLBH) 4;  # number of lines in parameter listbox
   # list of fixed word parametes (ligkern exceptional)
   set tfmpkGlob(fwParList) \
       {width height depth italic ligkern kern exten};

   foreach param $tfmpkGlob(fwParList) {
      set paramf [format "%sf" $param];
      set Param [capitalFirst $param];
      set paramsLabel [format "%ssLabel" $param];
      set tfmpkGlob($paramsLabel) $Param;
      append tfmpkGlob($paramsLabel) "s\[\]";

      set  paramsList [format "%ssList" $param];
      set tfmpkGlob($paramsList) {}

      set tfmpkGlob([format "n%ss" $Param]) "";
   }

   # preferences   
   set tfmpkGlob(maxTabGeo)  1;
   # 800x600 ??
   set tfmpkGlob(geomMaxTabLast) 600x400;
   set tfmpkGlob(geomMaxTab)     600x400;
}; # initGlobals


#############################
proc setFontPaths {} {
   global tfmpkGlob;
   #set tfmpkGlob(tfmPath)  ".:/usr/TeX/lib/texmf/fonts/public//tfm//"
   #set tfmpkGlob(pkPath)   ".:/usr/TeX/lib/texmf/fonts//pk//"
   if {[catch {set TEXMF [exec kpsexpand \$TEXMF]}]} {
      set TEXMF /usr/lib/texmf/texmf
   }
   if {[catch {set $VARFONTS [exec kpsexpand \$$VARFONTS]}]} {
      set VARFONTS /var/lib/texmf/fonts
   }
   foreach tp {tfm pk} {
      set tpPath [format "%sPath" $tp];
      set p .;
      foreach d [list $TEXMF/fonts $VARFONTS] {
         set dp $d/$tp;
         dputs "d=$d, dp=$dp";
         if {[file isdirectory $dp]} {
            set p [format "%s:%s//" $p $dp];
         }
      }
      set tfmpkGlob($tpPath) $p;
   }
   dputs "tfmPath = $tfmpkGlob(tfmPath)";
   dputs "pkPath  = $tfmpkGlob(pkPath)";

   env_override tfmpkGlob(tfmPath) "TFMFONTS"
   env_override tfmpkGlob(pkPath)  "PKFONTS"
}; # setfontpaths


#############################
# Set icon and location
proc setIconLoc {win iconWin iconName} {
   global tfmpkGlob;

   dputs "setIcon win=$win iconWin=$iconWin iconName=$iconName";
   dputs "bitmap=$tfmpkGlob(bitmapIcon)";

   wm iconname $win [format "TfmPk %s" $iconName];  # Default


   if {$tfmpkGlob(bitmapIcon)} {

      if {![info exists tfmpkGlob(iconBitmapFile)]} {
         set iconXbmFile [findFile tfmpk.xbm];
	 set tfmpkGlob(iconBitmapFile) $iconXbmFile;
         dputs "iconXbmFile=$iconXbmFile";
         if {![string compare $iconXbmFile ""]} {
            warn "Icon tfmpk.xbm file not found";
         }
      }

      if {[info exists tfmpkGlob(iconBitmapFile)]} {
         wm iconbitmap $win @$tfmpkGlob(iconBitmapFile);
         wm iconname $win $iconName;
      }

   } else {
      if {![info exists tfmpkGlob(iconImage)]} {
         set iconGif [findFile tfmpk.gif];
         dputs "iconGif=$iconGif";
         if {![string compare $iconGif ""]} {
            warn "Icon tfmpk.gif file not found";
         } else {
            set tfmpkGlob(iconImage) [image create photo  -file $iconGif];
         }
      }

      if {[info exists tfmpkGlob(iconImage)]} {
         if {![winfo exists $iconWin]} {
            toplevel $iconWin;
            label $iconWin.lbl -image $tfmpkGlob(iconImage);
            pack $iconWin.lbl;
         }
         wm iconwindow $win $iconWin;
         wm iconname $win $iconName;
         # wm iconname $iconWin iconName;
      }
   }

   if {[info exists tfmpkGlob(geoloc,$win)]} {
      dputs "geloc=$tfmpkGlob(geoloc,$win)"
      wm geometry $win $tfmpkGlob(geoloc,$win);
   }

}; # setIconLoc


########################################################################
# Fixed window with icon
########################################################################
proc winFixedIcon {w title iconName} {
   # #dpts "winFixedIcon: "
   #dpts "winFixedIcon: w=$w title=$title, iconName=$iconName";
   wm title $w "$title";
   wm resizable $w 0 0;
   set  wIcon  [format "%sicon" $w];
   #dpts "wIcon=$wIcon";
   setIconLoc $w $wIcon $iconName;
   return $w;
}; # winFixedIcon


########################################################################
# General top-level fixed window with icon
########################################################################
proc topLvlFixedIcon {wn title {iconName ""} args} {
   #dpts "tpLvlFixdIcon: wn=$wn title=$title, icnTitl=$iconName, args=$args";
   eval toplevel $wn $args;
   winFixedIcon $wn $title $iconName;
   dputs "checking form geloc"
   return $wn;
}; # topLvlFixedIcon


#############################
# Warning. Wait for User's OK
proc warn {msg {parent .}} {
   global tfmpkGlob;
   tk_messageBox -title "TfmPk Warning" \
      -parent $parent -icon warning -type ok -message $msg;
}; # warn


#############################
# Warning. Wait for User's OK
proc warn0 {msg} {
   global tfmpkGlob;
   set f [toplevel .warn -borderwidth 10];
   wm title $f "TfmPk - Warning";
   message $f.msg -aspect 600 -text $msg;
   pack $f.msg -side top -fill x;
   # button $f.ok -text OK ai-command {[list set tfmpkGlob(warnOK) 1]};
   button $f.ok -text OK -command {set tfmpkGlob(warnOK) 1};
   pack $f.ok -side bottom -fill x;
   focus $f.ok;
   grab $f.ok;
   tkwait variable tfmpkGlob(warnOK);
   grab release $f.ok;
   safeDestroy $f;
}; # warn0


########################################################################
# Return the commom directory length
# Examples:
#   [comdirlen "/usr/TeX/foo" "/usr/TeX/fubar"] returns  9. ("/usr/TeX/")
#   [comdirlen "quantity" "quality"] returns  0.
#
proc comdirlen {s1 s2} {
   # Force difference
   dputs "comdirlen $s1 $s2";
   append s1 X;
   append s2 Y;
   set com "";
   set l 0;
   set equal 1;
   set lastDirBdyInd 0;
   while {$equal} {
      set c1 [string index $s1 $l];
      set c2 [string index $s2 $l];
      set cmp [string compare $c1 $c2];
      if {[expr ($cmp == 0)]} {
         set cmp [string compare $c1 "/"];
         if {[expr ($cmp == 0)]} {
            set lastDirBdyInd $l
         }
         set l [expr $l+1];
      } else {
         set equal 0;
      }
   }
   dputs "lastDirBdyInd=$lastDirBdyInd";
   return $lastDirBdyInd;
}; # comdirlen


########################################################################
proc listboxSetlist {lb list} {
   global tfmpkGlob;
   dputs "YM listboxSetlist: lb=$lb, list=$list";
   # setXsizes list;
   $lb delete 0 end;
   eval {$lb insert 0} $list;
}; # listboxSetlist


########################################################################
# TFM general information
proc tfmGeneral {} {
   dputs "tfmGeneral";
   global tfmpkGlob;
   set f .tfmGeneral;
   if {![winfo exists $f]} {
      topLvlFixedIcon $f "TfmPk - TFM-General" TFM-Head \
                      -borderwidth 10 -bg $tfmpkGlob(color,TFM);
      message $f.msg -aspect 600 -textvariable tfmpkGlob(tfmGeneral);
      pack $f.msg -side top -fill x;
      button $f.close -text Close  -command [list safeDestroy $f];
      pack $f.close -side bottom -fill x;
   }
   wm deiconify $f;
   raise $f;
   set tfmpkGlob(tfmGeneral) [tfmpk tfminfo general];
}; # tfmGeneral


########################################################################
# PK general information
proc pkGeneral {} {
   dputs "pkGeneral";
   global tfmpkGlob;
   set f .pkGeneral;
   if {![winfo exists $f]} {
      topLvlFixedIcon $f "TfmPk - PK-General" PK-Head -borderwidth 10 -bg pink2;
      message $f.msg -aspect 600 -textvariable tfmpkGlob(pkGeneral);
      pack $f.msg -side top -fill x;
      button $f.close -text Close  -command [list safeDestroy $f];
      pack $f.close -side bottom -fill x;
   }
   wm deiconify $f;
   raise $f;
   set tfmpkGlob(pkGeneral) [tfmpk pkGeneral];
}; # tfmGeneral


########################################################################
# Valid regular expression - I wish TCL would do it without catch
########################################################################
proc regexpOK {re} {
   # puts "regexpOK re=$re";
   set ok 1;
   if {[catch {regexp -- $re any}]} {
      set ok 0;
   }
   return $ok;
}; # regexpOK


########################################################################
# check validity of regular expression. interrpet ending $
proc getFontPatternOld {} {
   global tfmpkGlob;
   set p $tfmpkGlob(namePattern);  dputs "p=$p";
   if {![string compare $p \$] ||![regexpOK $p]} {
      warn "Could not parse regular expression '$p'.";
      set p ""; # failure
   } else {
      if {![string compare $p ""]} {
         set p .+;
      } else {
         set l [string length $p];
         set lastChar [string range $p [expr $l-1] end];
         dputs "lastChar=$lastChar";
         if {![string compare $lastChar \$]} {
            set p [string range $p 0 [expr $l-2]];
         } else {
            set p [format "%s.*" $p ];
         }
      }
   }
   dputs "getFontPattern: p=$p";
   return $p;
}; # getFontPatternOld


########################################################################
# check validity of regular expressions. interrpet ending $
proc getFontPatterns {} {
   global tfmpkGlob;
   set p $tfmpkGlob(namePattern);  dputs "p=$p";
   set pDPI $tfmpkGlob(dpiPattern);  dputs "pDPI=$pDPI";
   if {![string compare $p \$] ||![regexpOK $p]} {
      warn "Could not parse regular expression '$p'.";
      set p ""; # failure
      set pDPI ""; # failure
   } else {
      if {![string compare $pDPI \$] ||![regexpOK $pDPI]} {
         warn "Could not parse regular expression '$pDPI'.";
         set p ""; # failure
         set pDPI ""; # failure
      } else {
         if {![string compare $pDPI ""]} {
            set pDPI "\[0-9\]+";
         }
         if {![string compare $p ""]} {
            set p .+;
         } else {
            set l [string length $p];
            set lastChar [string range $p [expr $l-1] end];
            dputs "lastChar=$lastChar";
            if {![string compare $lastChar \$]} {
               set p [string range $p 0 [expr $l-2]];
            } else {
               set p [format "%s.*" $p ];
            }
         }
      }
   }
   dputs "getFontPatterns: p=$p, pDPI=$pDPI";
   return [list $p $pDPI];
}; # getFontPatterns


########################################################################
proc listFonts {} {
   global tfmpkGlob debug;
   dputs listFonts;

   set pats [getFontPatterns];
   set pat [lindex $pats 0];
   set patDPI [lindex $pats 1];
   dputs "listFonts: pat=$pat, patDPI=$patDPI";
   if {[string compare $pat ""] && [string compare $patDPI ""]} {
      #  $pat\.tfm$   $pat\.[0-9]+pk$
      set re [format "%s\\.tfm$" $pat];  dputs "re=$re";
      set tfmList [lspath -regexp base $re $tfmpkGlob(tfmPath)];
      set re [format "%s\\.%spk$" $pat $patDPI];  dputs "re=$re";
      set pkList  [lspath -regexp base $re $tfmpkGlob(pkPath)];
      eval set list3 [eval tfmpk matchFonts \
                      [llength $tfmList] [llength $pkList] $tfmList $pkList];
      set ll3 [llength $list3];
      dputs "list3=$list3, #=$ll3";
      set tfmpkGlob(fontList) [lindex $list3 0];
      set tfmpkGlob(tfmList) [lindex $list3 1];
      set tfmpkGlob(pkList) [lindex $list3 2];
      if {$debug} {
         foreach kl {fontList tfmList pkList} {
            set l $tfmpkGlob($kl);   set ll [llength $l];
            dputs "kl=$kl, ll=$ll, l=$l";
         }
      }

      set lb $tfmpkGlob(fontListbox);
      $lb delete 0 end;
      eval {$lb insert 0} $tfmpkGlob(fontList);
   }
}; # listFonts


########################################################################
proc listSelectTfmPk {ix ff} {
   global tfmpkGlob;
   if {$ix >= 0} {
      set tfmpkGlob(etfm) [lindex $tfmpkGlob(tfmList) $ix];
      set tfmpkGlob(epk)  [lindex $tfmpkGlob(pkList) $ix];
      set tfmpkGlob(ldfntMsg2) "";
      $ff.etfm xview end;
      $ff.epk xview end;
   }
}; # listSelectTfmPk


########################################################################
proc listSelectFont {y ff} {
   global tfmpkGlob;
   set ix [$tfmpkGlob(fontListbox) nearest $y];
   if {$ix >= 0} {
      set font [$tfmpkGlob(fontListbox) get $ix];
      focus $tfmpkGlob(fontListbox);
      dputs "ix=$ix, font=$font";
      listSelectTfmPk $ix $ff;
   }
}; # listSelectFont


########################################################################
proc listFontKeySeek {lb c} {
   global tfmpkGlob;
   set iSeek -1;
   set seekstr [format %s%s $tfmpkGlob(ldfntMsg2) $c];
   dputs "seekstr=$seekstr";
   if {![string compare $c \b]} { # backspace
      set l [string length $seekstr];
      set seekstr [string range $seekstr 0 [expr $l-3]];
      dputs "backspace l=$l seekstr=$seekstr";
      if {[string length $seekstr] == 0} {
         set iSeek 0;
      }
   }
   if {$iSeek == -1} {
      # binary search
      set low 0; 
      set s [$lb size];
      set high $s;
      while {$low < $high} {
         set m [expr ($low+$high)/2];
         set fn [$lb get $m];
         set cmp [string compare $seekstr $fn];
         dputs "low=$low m=$m high=$high fn=$fn, cmp=$cmp";
         if {$cmp == 0} {
            set low  $m;
            set high $m;
         } elseif {$cmp < 0} {
            set high $m;
         } else {
            set low [expr $m+1];
         }
      }
      if {$low < $s} {
         set fn [$lb get $low];
         dputs "seekstr=$seekstr, fn=$fn"
         if {[string first $seekstr $fn] == 0} {
            set iSeek $low;
         }
      }
   }
   dputs "iSeek=$iSeek";
   if {$iSeek >= 0} {
      set tfmpkGlob(ldfntMsg2) $seekstr;
      $lb see $iSeek;
      $lb selection clear 0 end;
      $lb selection set $iSeek;
      $lb selection anchor $iSeek;
      $lb activate $iSeek;
      set tfmpkGlob(etfm) [lindex $tfmpkGlob(tfmList) $iSeek];
      set tfmpkGlob(epk)  [lindex $tfmpkGlob(pkList) $iSeek];
   } else {
      bell;
   }
}; # listFontKeySeek

########################################################################
proc listFontSeek {op lb args} {
   global tfmpkGlob;
   dputs "args=$args"
   if {![string compare $op FocusIn]} {
      focus $lb;
      set tfmpkGlob(ldfntMsg1) "Type to seek:";
      set tfmpkGlob(ldfntMsg2) "";
   } elseif {![string compare $op FocusOut]} {
      set tfmpkGlob(ldfntMsg1) "";
      set tfmpkGlob(ldfntMsg2) "";
   }  elseif {![string compare $op keyRel]} {
      set c [lindex $args 1];
      set cl [string length $c];  dputs "cl=$cl";
      if {$cl == 1} {
         listFontKeySeek $lb $c;
      } else {
         set f [lindex $args 0];
         set iSel [$lb curselection];  dputs "iSel=$iSel, f=$args";
         listSelectTfmPk $iSel $f;
      }
   }
}; # listFontSeek


########################################################################
proc checkCheckSums {} {
   set CSEQ "Check-Sum = ";
   set lCSEQ [string length $CSEQ];
   foreach inf {tfminfo pkGeneral} {
      set tfmGenInfo [tfmpk $inf general];
      dputs "tfmGenInfo=$tfmGenInfo";
      set pCSEQ [string first $CSEQ $tfmGenInfo];
      set cs($inf) [string range $tfmGenInfo [expr $pCSEQ+$lCSEQ] end];
      dputs "inf=$inf cs=$cs($inf)";
   }
   if {[string compare $cs(tfminfo) $cs(pkGeneral)]} {
      warn "Unmatched check-sums:\nTFM = $cs(tfminfo)\nPK  = $cs(pkGeneral)";
   }
}; # checkCheckSums


########################################################################
proc syncWidgetsState {} {
   global tfmpkGlob;
   set state [sifthenelse $tfmpkGlob(dataLoaded) normal disabled];
   .files.butHdrTFM configure -state $state;
   .files.butHdrPK  configure -state $state;
   $tfmpkGlob(fileMenu) entryconfigure $tfmpkGlob(reopIndex) -state $state;
   $tfmpkGlob(fileMenu) entryconfigure $tfmpkGlob(unloadIndex) -state $state;
}; # syncWidgetsState


########################################################################
proc unload {} {
   global tfmpkGlob;
   set tfmpkGlob(dataLoaded) 0;
   safeDestroy .tf;
   safeDestroy .focusChar;
   safeDestroy .extenShow;
   safeDestroy .kernShow;
   set tfmpkGlob(comdir) "";
   set tfmpkGlob(currentTFM) "";
   set tfmpkGlob(currentPK) "";
   foreach param $tfmpkGlob(fwParList) {
      set paramf [format "%sf" $param];
      set f $tfmpkGlob($paramf);
      $f.ind delete 0 end;
      $f.vl delete 0 end;
   }
   syncWidgetsState;
}; # unload


########################################################################
proc loadTfmPk {{destroyMe nOthIng}} {
   global tfmpkGlob;
   dputs loadTfmPk

   unload;
   set tfmfn $tfmpkGlob(etfm);
   set pkfn $tfmpkGlob(epk);
   set ok [expr  [file exists $tfmfn] && [file exists $pkfn]];

   set result [tfmpk loadPK $pkfn 0]; # 255 for full debug
   dputs "loadPK= $result";
   if {[string length $result]} {
      warn $result;
      set ok 0;
   }

   set result [tfmpk loadTFM $tfmfn 0]; # 255 for full debug
   dputs "loadTFM= $result";
   if {[string length $result]} {
      warn $result;
      set ok 0;
   }

   # Need to match check-sum!! LATER
   set tfmpkGlob(dataLoaded) $ok;
   syncWidgetsState;
   if {$ok} {
      if {($tfmpkGlob(flags) & 1)} {
         set sizeOK [tfmpk tfminfo sizeOK];
         if {$sizeOK != 1} {
            warn "Unexpected TFM-file size";
         }
      }
      set cml [comdirlen $tfmfn $pkfn];
      if {$cml <= 0} {
         set tfmpkGlob(comdir) "";
         set tfmpkGlob(currentTFM) $tfmfn;
         set tfmpkGlob(currentPK)  $pkfn
      } else {
         set cmlm1 [expr $cml-1];
         set cmlp1 [expr $cml+1];
         set tfmpkGlob(comdir) [string range $tfmfn 0 $cmlm1];
         set tfmpkGlob(currentTFM) [string range $tfmfn $cmlp1 end];
         set tfmpkGlob(currentPK)  [string range $pkfn $cmlp1 end];
      }

      safeDestroy $destroyMe;
      set tfmpkGlob(lastKernIndex) -1;
      checkCheckSums
      show;
   }

   # Now stretch/shrink entries
   set le [string length $tfmpkGlob(currentTFM)];  
   .files.tfme configure -width [maximize le 16];
   set le [string length $tfmpkGlob(currentPK)];
   .files.pke configure -width [maximize le 16];

}; # loadTfmPk


########################################################################
proc selectTFMPK {typ enm} {
   global tfmpkGlob tk_strictMotif;
   set saveMotif $tk_strictMotif;
   set tk_strictMotif 1;
   set fn [tk_getOpenFile -parent . -title "Select $typ file"];
   if {[string compare $fn ""]} {
      set tfmpkGlob($enm) $fn;
   }
   set tk_strictMotif $saveMotif
}; # selectTFMPK


########################################################################
# Dialog to choose a font, i.e.  <foo>.tfm <foo>.<bar>pk files to load
proc loadFont {} {
   global tfmpkGlob;
   dputs loadFont;
   set w .loadFont
   if {![winfo exists $w]} {
      topLvlFixedIcon $w "TfmPk - Load Font" Load -borderwidth 10;
      wm resizable $w 1 0;    # but make its width resizable
      set lbl [label $w.l -text "Load Font"];
      pack $lbl -side top;

      # explicit file names
      dputs "explicit file names";
      set f [frame $w.f1];
      pack $f -side top -expand true -fill x;
      set buttfm [button $f.buttfm -text ".tfm file..." \
               -bg $tfmpkGlob(color,TFM) -command [list selectTFMPK TFM etfm]];
      set etfm [entry $f.etfm -textvariable tfmpkGlob(etfm) \
               -bg $tfmpkGlob(color,TFM)];
      $etfm xview end;
      set butpk [button $f.lblpk -text ".pk file..." -bg pink2 \
                 -command [list selectTFMPK PK epk]];
      set epk [entry $f.epk -textvariable tfmpkGlob(epk) -bg pink2];
      $epk xview end;
      grid $buttfm -row 1 -sticky ew;    grid $etfm -row 1 -column 2 -sticky ew;
      grid $butpk -row 2 -sticky ew;     grid $epk  -row 2 -column 2 -sticky ew;
      grid columnconfigure $f 2 -weight 99;

      # List & Selection
      dputs "List & Selection";
      set fls [frame  $w.fls];
      pack $fls -side top -pady 10 -expand true -fill x;
         pack [label $fls.lbl -text "List & Select"] -side top;
         set fe [frame $fls.fe];
         pack $fe -side top -expand true -fill x -pady 5;
            set buttfm [label $fe.buttfm -text "TFM Path" \
                    -bg $tfmpkGlob(color,TFM)];
            set etfm [entry $fe.etfm -textvariable tfmpkGlob(tfmPath) \
                      -bg $tfmpkGlob(color,TFM)];
            set butpk [label $fe.lblpk -text "PK Path" -bg pink2];
            set epk [entry $fe.epk -textvariable tfmpkGlob(pkPath) -bg pink2];
            grid $buttfm $etfm -sticky ew;
            grid $butpk $epk -sticky ew;
            grid columnconfigure $fe 1 -weight 99;
         set fl [frame $fls.fl];
         pack $fl -side top;
         set fblist [frame $fl.fblist];

            set ffblist [frame $fblist.f]
            pack $ffblist -side top -pady 3; 
            pack [label $ffblist.l -text "Name RegExp:"] -side top;
            pack [entry $ffblist.e -textvariable tfmpkGlob(namePattern) \
                  -width 8] -side top;
              bind $ffblist.e  <FocusIn> [list set tfmpkGlob(ldfntMsg1) \
               "Regular expression to filter next (Re)List"];
            bind $ffblist.e <Return> listFonts;
            bind $ffblist.e  <FocusOut> [list set tfmpkGlob(ldfntMsg1) ""];
            bind $ffblist.e <Key-F1> [list help NameRegExp %#];

            set ffblist [frame $fblist.fdpi]
            pack $ffblist -side top -pady 3; 
            pack [label $ffblist.l -text "DPI RegExp:"] -side top;
            pack [entry $ffblist.e -textvariable tfmpkGlob(dpiPattern) \
                  -width 8] -side top;
              bind $ffblist.e  <FocusIn> [list set tfmpkGlob(ldfntMsg1) \
               "Resolution RegExp to filter next (Re)List"];
            bind $ffblist.e <Return> listFonts;
            bind $ffblist.e  <FocusOut> [list set tfmpkGlob(ldfntMsg1) ""];
            bind $ffblist.e <Key-F1> [list help DPIRegExp %#];

            set blist [button $fblist.blist -text "(Re)\nList" \
                      -command listFonts];
            pack $blist -side bottom -pady 6;
            bind $blist <Key-F1> [list help relist %#];
         pack $fblist -side left -padx 6;
         set lb [listbox $fl.list -setgrid true -selectmode browse \
                 -height 8 \
                 -yscrollcommand [list $fl.scroll set]];
         set sy [scrollbar $fl.scroll -command [list $lb yview]];
         pack $sy -side right -fill y;
         pack $lb -side left -fill both -expand true;
         set tfmpkGlob(fontListbox) $lb;
         bind $lb <ButtonRelease-1> [list listSelectFont %y $f];
         bind $lb <Double-Button-1> loadTfmPk;
         bind $lb <Key-F1> [list help listselect %#];
         # bind $lb <Enter> [list focus $lb];
         bind $lb <FocusIn> [list listFontSeek FocusIn $lb];
         bind $lb <FocusOut> [list listFontSeek FocusOut $lb];
         # bind $lb <Leave> [list listFontSeek leave $lb];
         # bind $lb <Key> [list listFontSeek  key $lb %A];
         bind $lb <KeyRelease> [list listFontSeek  keyRel $lb $f %A];
         eval {$lb insert 0} $tfmpkGlob(fontList)
         set tfmpkGlob(fontListBox) $lb;

      dputs Bottom;
      set fmsg [frame $w.msg];
      pack $fmsg -side bottom -anchor w;
      set tfmpkGlob(ldfntMsg1) "";
      set tfmpkGlob(ldfntMsg2) "";
      pack [label $fmsg.l1 -textvariable tfmpkGlob(ldfntMsg1)] -side left;
      pack [label $fmsg.l2 -textvariable tfmpkGlob(ldfntMsg2) \
            -foreground darkgreen] -side left;

      set fbot [frame $w.fbot];
      pack $fbot -side bottom;
      set load [button $fbot.load -text Load -width 5 -command loadTfmPk];
      set ok [button $fbot.ok -text OK -width 5 -command [list loadTfmPk $w]];
      set close [button $fbot.close -text Close -command [list safeDestroy $w]];
      grid $load $ok $close -sticky ew -padx 8

      bind $w <Key-F1> [list help loadFont %#];
   }
   wm deiconify $w;
   raise $w;
   focus $tfmpkGlob(fontListBox);
   dputs "loadFont: end proc";
}; # loadFont


########################################################################
proc tfmpkExit {} {
   dputs tfmpkExit;
   destroy .
}; # tfmpkExit


########################################################################
proc tfmParamYView {fparam args} {
   #dputs "tfmParamYView fparam=$fparam args=$args";
   eval $fparam.ind yview $args;
   eval $fparam.vl yview $args;
}; # tfmParamYView


########################################################################
proc tfmlb {param} {
   dputs "tfmlb $param"
   global tfmpkGlob;
   append paramf $param f;
   append paramLBflag $param LBflag;
   set f $tfmpkGlob($paramf);

   if {![winfo exists $f.vl]} {
       set cursor top_left_arrow; # default
       #  -exportSelection false
       #  -exportselection
       set w 7;
       if {![string compare $param ligkern]} {
          set w 18
          set cursor bogosity
       };
       if {![string compare $param kern]} {set w 9};
       if {![string compare $param exten]} {
          set w 15
          set cursor double_arrow;
       };
       listbox $f.ind -width 3 -height 8 \
         -font fixed -cursor $cursor \
         -exportselection false \
         -yscrollcommand [list $f.sy set];
       listbox $f.vl -width $w -height 8 \
         -font fixed -cursor $cursor \
         -exportselection false \
         -yscrollcommand [list $f.sy set];
      bind $f.ind <ButtonRelease-1> [list showParamIndex $param %W %x %y];
      bind $f.vl <ButtonRelease-1> [list showParamIndex $param %W %x %y];
      if {[lsearch {ligkern exten} $param] >= 0} {
         bind $f.vl <Enter> [list set tfmpkGlob(mainMessage) \
                            "Click on <xNN> for character focus"];
         bind $f.vl <Leave> [list set tfmpkGlob(mainMessage) ""];
      }
      scrollbar $f.sy -orient vertical -command [list tfmParamYView $f];
   }
   pack $f.sy -side right -fill y;
   pack $f.ind -side left;
   pack $f.vl -side left;
   append paramsList $param sList;
   listboxSetlist $f.vl $tfmpkGlob($paramsList)
   set ll [llength $tfmpkGlob($paramsList)];
   $f.ind delete 0 end;
   for {set i 0} {$i < $ll} {incr i} {
      $f.ind insert $i $i;
   }
}; # tfmlb


########################################################################
proc tfmFrameParam {frm param} {
   global  tfmpkGlob;
   set f [frame $frm.$param -borderwidth 4];
   pack $f -side left;
   append paramf $param f;
   dputs "tfmFrameParam $param, f=$f, paramf=$paramf";
   set tfmpkGlob($paramf) $f;
   append paramsLabel $param sLabel;
   label $f.cb -textvariable tfmpkGlob($paramsLabel);
   pack $f.cb -side top;
   tfmlb $param;
}; # tfmFrameParam


########################################################################
proc charMetric {} {
   global tfmpkGlob;
   dputs charMetric;

   pack [frame .tfminfo] -side top ;
   set tf1 [frame .tfminfo.tfmiWHDI];  pack $tf1 -side top;

   foreach param {width height depth italic} {
      tfmFrameParam $tf1 $param;
   }
   set tf2 [frame .tfminfo.tfmiLKE];   pack $tf2 -side top;
   foreach param {ligkern kern exten} {
      tfmFrameParam $tf2 $param;
   }
}; # charMetric


########################################################################
# Font Char - Probe
proc focusCharProbe {mode w {x -1} {y -1}} {
   global tfmpkGlob;
   # dputs "focusCharProbe $mode $x $y";
   if {[string compare $mode leave]} {
      set msg [tfmpk focusCharProbe $x $y];
   } else {
      set msg "Use pointer to probe";
   }
   set oldWidth [$w cget -width];
   set newWidth [string length $msg];
   if {$newWidth < $oldWidth} {
      set newWidth $oldWidth;
   }
   $w configure -text $msg -width $newWidth;
}; # focusCharProbe


########################################################################
# focus grid on/off callback
proc focusGridUpdate {} {
   global tfmpkGlob;
   dputs "focusGridUpdate: $tfmpkGlob(focusGridOn)";
   tfmpk focusCharGrid $tfmpkGlob(focusGridOn);
   set cmd [list $tfmpkGlob(winfc) redraw];
   eval $cmd;
}; # focusGridUpdate

########################################################################
# Focus Char - window
proc focusCharWindow {} {
   global tfmpkGlob debug;
   set w .focusChar;
   if {![winfo exists $w]} {
      topLvlFixedIcon $w "TfmPk Focus Character" Focus -cursor cross;
      label $w.lab -textvariable tfmpkGlob(fflab);
      pack $w.lab -side top;

      set tfmpkGlob(winfc) [ximple $w.focusChar -cursor crosshair];
      pack $w.focusChar -side top -padx 5 -pady 5;
      # set tfmpkGlob(focusCharProbe) "Use pointer to probe";

      label $w.tfminfo -textvariable tfmpkGlob(focusTFMinfo) \
            -bg $tfmpkGlob(color,TFM);
      pack $w.tfminfo -side top -padx 3;
      bind $w.tfminfo <Enter> [list $w.probe configure -width 30 \
                              -text "TFM indices (see main Window)"];
      bind $w.tfminfo <Leave> [list focusCharProbe leave $w.probe];

      set f [frame $w.pkinf -bg pink2];
      pack $f -side top -padx 3;
      set msg(BM) "Bitmap Size";
      # set msg(HV) "Horizontal & Vertical Offset";
      set msg(HV) "X,Y Offset";
      set msg(DX) "Pixel dx-Escapement";
      foreach e {BM HV DX} {
         set key [format "focusPKinf%s" $e];
         set lbl [label $f.l$e -textvariable tfmpkGlob($key) -bg pink2];
         pack $lbl -side left -padx 4;
         bind $lbl <Enter> [list $w.probe configure -text $msg($e)];
         bind $lbl <Leave> [list focusCharProbe leave $w.probe];
         # bind $lbl <Enter> [list dputs "e=$e"];
      }

      checkbutton $w.grid -text Grid -variable tfmpkGlob(focusGridOn) \
         -command focusGridUpdate;
      pack $w.grid -side top -anchor w -padx 10 -pady 4;

      label $w.probe -width 1; #  -textvariable tfmpkGlob(focusCharProbe);
      pack $w.probe -side top -pady 5;
      bind $w.focusChar <Motion> [list focusCharProbe move $w.probe %x %y];
      bind $w.focusChar <Leave> [list focusCharProbe leave $w.probe];
      focusCharProbe leave $w.probe; # init probe-message as if we left
      button $w.close -text Close -command [list safeDestroy $w];
      pack  $w.close -side bottom -pady 8;
   }
}; # focusCharWindow


########################################################################
# About text
########################################################################
proc about {} {
   global tfmpkGlob;
   set winName .about;
   if {[winfo exists $winName]} {
      #dpts "raising winName=$winName";
      wm deiconify $winName;
      raise $winName;
   } else {
      topLvlFixedIcon $winName "TFMPK - About" About;

      button $winName.close -text Close -command [list safeDestroy $winName];
      pack $winName.close -side bottom -pady 6;

      label $winName.info -bitmap info;
      pack $winName.info -side left -padx 6;

      set t [text $winName.text -setgrid true -wrap word \
             -font *-times-medium-r-*-12-* \
            -width 56 -height 14 -padx 6];
      pack $t -side left  -padx 6 -pady 6;

      $t tag configure boldBig -font *-times-bold-r-*-14-*
      $t tag configure fixed -font fixed

      $t insert end  "              About TFMPK $tfmpkGlob(version)\n" boldBig;
      $t insert end  \
"\nTFMPK is a TeX (.tfm+.pk) font viewer. \n\n";
      $t insert end \
         "TFMPK uses (and rhymes with) Tcl/Tk which is available from:\n"
      $t insert end /anonymous@ftp.smli.com:/pub/tcl/" fixed;
      $t insert end \
"\n\nTFMPK is written by Yotam Medini (yotam@blueneptune.com).\n" fixed;
      $t insert end  "It is available from:\n";
      $t insert end \
"  /anonymous@ftp.blueneptune.com:/pub/users/yotam/\n" fixed;
      $t insert end  "\n";
      $t insert end  "See also: http://www.blueneptune.com/~yotam/tfmpk.html";
      $t configure -state disabled;
   }
}; # about


########################################################################
# Main Window + ALso create the Focus-Char window
proc mainWindow {} {
   global tfmpkGlob debug;

   # Set low priority options.
   option add *backround    LightGray      widgetDefault
   option add *forground    black          widgetDefault
   option add *borderColor  brown          widgetDefault
   option add *textColor    DarkSlateGrey  widgetDefault
   option add *selectColor  DarkOliveGreen widgetDefault

   # Fonts
   option add Tfmpk*font            -Adobe-Helvetica-Bold-R-Normal--*-120-*
   option add Tfmpk*Entry*font      -*-*-bold-r-*--*-120-*-*-m-*
   option add Tfmpk*Menubutton*font -Adobe-Helvetica-Bold-O-Normal--*-140-*
   option add Tfmpk*Listbox*font    fixed


   # Create a frame and buttons along the top
   set mw .;
   bind $mw <Key-F1> [list help general %#];
   winFixedIcon $mw "TFM-PK - Main" Main

   set mb [frame $mw.buttons -highlightthickness 1 -relief raised -bd 4];
   pack $mb -side top -fill x

   ###########
   # File menu
   ###########
   # -font $hdfvGlob(boldFont)
   set f [menubutton $mb.file -text File -menu $mb.file.menu];
   pack $f -side left -padx 10 -pady 6;
   set m [menu $mb.file.menu -tearoff 0];
   set tfmpkGlob(fileMenu) $m;
   $m add command -label Open... -command loadFont;

   $m add command -label "Reopen Current" -command loadTfmPk -state disabled;
   set tfmpkGlob(reopIndex) 1;
   $m add command -label "Unload" -command unload -state disabled;
   set tfmpkGlob(unloadIndex) 2;

   $m add separator;
   $m add command -label Exit -command tfmpkExit;

   if {$debug} {
      $m add separator;
      $m add command -label "Reload Tcl-code" -command [list source tfmpk.tcl];
   }

   ##############
   # Options menu
   ##############
   set f [menubutton $mb.options -text Options -menu $mb.options.menu];
   pack $f -side left -padx 10 -pady 6;
   set m [menu $mb.options.menu -tearoff 0];
   $m add command -label Preferences... -command editPref;

   # button .buttons.pref -text Preferences -command editPref;
   # pack .buttons.quit .buttons.reset .buttons.pref -side right;

   ##############
   # Windows menu
   ##############
   set f [menubutton $mb.windows -text Windows -menu $mb.windows.menu];
   pack $f -side left -padx 10 -pady 6;
   set m [menu $mb.windows.menu -tearoff 0];
   $m add command -label "Font Table" -command bringUpTableWin;
   $m add command -label "Focus Char" -command bringUpFocusChar;


   ##########
   set help [menubutton $mb.help -text Help -menu $mb.help.menu];
   pack $help -side right -padx 10 -pady 6;
   set m [menu $help.menu -tearoff 0];
   $m add command -label About... -command about;
   $m add command -label General... -command [list help general -1];



   # TFM+PK files
   frame .files;       pack .files -side top -fill x
      label .files.comdirl -text "Common DIR:" -padx 0;
      label .files.comdire -textvariable tfmpkGlob(comdir);
   foreach tp {TFM PK} {
      set tpl [string tolower $tp];
      set b [format ".files.butHdr%s" $tp];
      set currtp [format "current%s" $tp];
      set cmd [format "%sGeneral" $tpl];
      button $b -text "$tp-File:" -bg $tfmpkGlob(color,$tp) -width 9 \
             -state disabled -command $cmd;
      bind $b <Enter> [list set tfmpkGlob(mainMessage) "Click for $tp header"];
      bind $b <Leave> [list set tfmpkGlob(mainMessage) ""];
      label [format ".files.%se" $tpl] -textvariable tfmpkGlob($currtp) \
            -bg $tfmpkGlob(color,$tp) -width 16;
   }


      grid .files.comdirl   .files.comdire -padx 5;
      grid .files.butHdrTFM .files.tfme -padx 5;
      grid .files.butHdrPK  .files.pke  -padx 5;

   # focusCharWindow;

   # Clicking on an item displays the font files
   # bind $tfmpkGlob(listbox) <ButtonRelease-1> \
   #   [list SelectFont $tfmpkGlob(listbox) %y]

   charMetric;

   set tools [frame .tfminfo.tools];  pack $tools -side bottom -pady 3;
   set ks [button $tools.kernShow -text KernShow -command kernShowStart];
   pack $ks -side left -padx 15;
   set tfmpkGlob(kernShowButton) $ks;
   set exs [button $tools.extenShow -text ExtenShow -command extenShowOpen];
   pack $exs -side right -padx 15;
   set tfmpkGlob(extenShowButton) $exs;

   set tfmpkGlob(mainMessage) "";
   pack [label $mw.msg -textvariable tfmpkGlob(mainMessage)] \
           -side bottom -fill x -anchor w;
}; # mainWindow


########################################################################
proc selectParamIndex {param i} {
   dputs "selectParamIndex $param $i";
   global tfmpkGlob;
   set paramf [format "%sf" $param];
   set f $tfmpkGlob($paramf);
   foreach iv {ind vl} {
      set lb $f.$iv;
      $lb selection clear 0 end;
      if {$i >= 0} {
         $lb see $i;
         $lb selection set $i;
         $lb selection anchor $i;
         $lb activate $i;
      }
   }
   # synchronize with index
   set yvvl [$f.vl yview];
   set yvvl0 [lindex $yvvl 0];
   $f.ind yview moveto $yvvl0;
   dputs "yvvl=$yvvl yvvl0=$yvvl0";
}; # selectParamIndex


########################################################################
#
proc showParamIndex {param w x y} {
   global tfmpkGlob;
   set n [$w nearest $y];

   update;
   selectParamIndex $param $n;

   set ww [$w cget -width]
   set www [winfo width $w]
   set ci -1;
   if {[string last .vl $w] >= 0} {
      set ci [expr ($x*$ww)/$www];
   }
   dputs "showParamIndex w=$w, x=$x y=$y, n=$n, ww=$ww, www=$www, ci=$ci";
   set cpick "";
   if {![string compare $param ligkern]} {
      set lkEntry [lindex $tfmpkGlob(ligkernsList) $n];
      # set kpos [string first K $lkEntry];  # xcc: xcc ff> K nn
      set kpos [string first "> K " $lkEntry];  # xcc: xcc ff> K nn
      if {$kpos > 0} {
         set iKern  [string range $lkEntry [expr $kpos+4] end]
         selectParamIndex kern $iKern;

         update; # Let entry $n be selected by the listbox widget
         ifKernShow;
      }
      if {(0 <= $ci) && ($ci <= 2)} {
         set cpick [string range $lkEntry 0 2];
      } elseif {(5 <= $ci) && ($ci <= 7)} {
         set cpick [string range $lkEntry 5 7];
      } elseif {(14 <= $ci) && ($ci <= 17)} {
         set cpick [string range $lkEntry 15 17];
      }
   } elseif {![string compare $param exten]} {
      extenShowUpdate;
      if {$ci >= 0} {
         set extenEntry [lindex $tfmpkGlob(extensList) $n];
         set ci4 [expr $ci - ($ci % 4)];
         dputs "extenEntry=$extenEntry, ci4=%ci4";
         set cpick [string range $extenEntry $ci4 [expr $ci4+2]];
      }
   }
   set rs [scan $cpick "x%x" c];
   dputs "cpick=$cpick, rs=$rs";
   if {$rs == 1} {
      dputs "c=$c";
      pickChar1 $c;
   }
}; # showParamIndex


########################################################################
proc safeCurselection {wlb} {
   # wlb get <index>
   set cs -1;
   if {[$wlb size] > 0} {
      set cs [$wlb curselection];
      if {[string length $cs] == 0} {
         set cs 0;
      }
   }
   return $cs;
}; # safeCurselection


########################################################################
proc currentKernIndex {} {
   global tfmpkGlob;
   set f $tfmpkGlob(ligkernf);
   set ki [safeCurselection $f.vl];  # ki = ligkern-entry-index
   if {$ki >= 0} {
      set lkEntry [$f.vl get $ki];
      dputs "lkEntry=$lkEntry";
      set kpos [string first "> K " $lkEntry];  # xcc: xcc ff> K nn
      if {$kpos < 0} {
         set ki $tfmpkGlob(lastKernIndex);
         if {$ki < 0} {
            set nk [$f.vl size];
            while {($kpos < 0) && ([incr ki] < $nk)} {
               set lkEntry [$f.vl get $ki];
               dputs "ki=$ki, lkEntry=$lkEntry";
               set kpos [string first K $lkEntry];  # xcc: xcc ff> K nn
            }
            if {$kpos < 0} {
               set ki -1;
            }
         }
      } else {
         set tfmpkGlob(lastKernIndex) $ki;
      }
   }
   dputs "currentKernIndex=$ki";
   return  $ki; # ki = ligkern-entry-index
}; # currentKernIndex


########################################################################
proc doKernShow {} {
   global tfmpkGlob;
   dputs doKernShow;
   # Tell kerning option
   set f $tfmpkGlob(kernf);
   set ki [currentKernIndex];  dputs "doKernShow: ki=$ki";
   if {$ki >= 0} {
      set ownerIndex 0;  # first character - owner index
      set kernOwnerLB $tfmpkGlob(kernOwnerLB);
      if {[winfo exists $kernOwnerLB]} {
         dputs "kernOwnerLB=$kernOwnerLB"
         set ownerIndex [safeCurselection $kernOwnerLB];
      }
      set owners $tfmpkGlob(kernOwners);
      set owner [lindex $owners $ownerIndex];
      dputs "ki=$ki, owners=$owners, oi=$ownerIndex, owner=$owner";

      # set the label
      set ownerCode [format "0x%02x" $owner];
      set f $tfmpkGlob(ligkernf);
      # set ki [safeCurselection $f.vl];  # ki = ligkern-entry-index
      set lkEntry [$f.vl get $ki];
      set kernedCode [string range $lkEntry 5 7];
      set tfmpkGlob(kernShowLabel) \
        [format "Kern: %s + 0%s" $ownerCode $kernedCode];

      tfmpk kernShow $ki $ownerIndex $tfmpkGlob(kernon);
      $tfmpkGlob(ximpleKernShow) redraw;
   }
}; # doKernShow


########################################################################
# configure kern-show window-size
proc kernShowWinSize {} {
   global tfmpkGlob;
   set w $tfmpkGlob(ximpleKernShow);
   dputs "kernShowWinSize w=$w";
   if {[winfo exists .kernShow]} {
      set result [tfmpk kernShowWinSize];
      scan $result "%d%d" width height;
      dputs "kernShowWinSize: $width $height";
      set cmd [list $w configure -borderwidth 2 -width $width -height $height];
      eval $cmd;
   }
}; # kernShowWinSize


########################################################################
set tfmpkGlob(kernon)  1;  # Global default
set tfmpkGlob(kernOwnerLB) 0;
########################################################################
proc kernShow {} {
   global tfmpkGlob;
   dputs "kernShow";
   set ftop .kernShow.top;
   if {![winfo exists .kernShow]} {
      dputs "Create .kernShow";
      topLvlFixedIcon .kernShow "TfmPk - Kern Show" Kern;

      set fbot [frame .kernShow.bot];
      pack $fbot -side bottom;

         checkbutton $fbot.kernon \
            -text Kern -variable tfmpkGlob(kernon) -command doKernShow;
         button $fbot.close -text Close -command [list safeDestroy .kernShow];
         pack $fbot.kernon $fbot.close -padx 15 -pady 10 -side left;

      frame $ftop;
      pack $ftop -side top -padx 4 -pady 3;

         set fshow [frame $ftop.show];
         pack $fshow -side right;


            set lab [label $fshow.lab -textvariable tfmpkGlob(kernShowLabel)];
            pack $lab -side top;

            set fx [frame $fshow.fx -bg black -borderwidth 2];
            pack $fx -side top;

            set tfmpkGlob(ximpleKernShow) [ximple $fx.kern];
            pack $tfmpkGlob(ximpleKernShow) -side top;
            kernShowWinSize;
      update;
   }

   set flb $ftop.flb;
   dputs "flb=$flb";
   if {[winfo exists $flb]} {
      dputs "destroy flb=$flb";
      destroy $flb;
   }

   set ki [currentKernIndex];
   set owners [tfmpk tfminfo ligkernOwners $ki];
   set tfmpkGlob(kernOwners) $owners;
   dputs "owners=$owners";
   if {[llength $owners] > 1} {
      dputs "Making listbox $flb.list"
      frame $flb;
      pack $flb -side left -padx 5;
      pack [label $flb.lbl -text 1st-Char] -side top;
      set tfmpkGlob(kernOwnerLB) [ \
         listbox $flb.list -setgrid true -selectmode browse -width 7 -height 5 \
            -yscrollcommand [list $flb.scroll set] ];
      dputs "tfmpkGlob(kernOwnerLB)=$tfmpkGlob(kernOwnerLB)";
      scrollbar $flb.scroll -orient vertical -command [list $flb.list yview];
      pack $flb.scroll -side right -fill y
      pack $flb.list -side left -fill both -expand true
      # bind $flb.list <ButtonRelease-1> {puts [list %W %y]};
      bind $flb.list <ButtonRelease-1> doKernShow;
      $flb.list delete 0 end;
      foreach owner $owners {
         set e [format "0x%02x" $owner];
         if {$owner > 0x20} {
            set e [format "%s %c" $e $owner];
         }
         $flb.list insert end $e;
      }
      $flb.list selection set 0;
   }
   doKernShow;
   dputs "kernShow  Done"
}; # kernShow


########################################################################
proc kernShowUpdate {} {
   global tfmpkGlob;
   dputs "kernShowUpdate";
   set vl [format "%s%s" $tfmpkGlob(ligkernf) .vl];
   dputs $vl;
   # set lki [$vl curselection];
   # puts "lki=$lki";
   kernShow;
}; # kernShowUpdate


########################################################################
proc kernShowStart {} {
   dputs "kernShowStart II";
   kernShowUpdate;
   raise .kernShow
}; # kernShowStart


########################################################################
proc ifKernShow {} {
   dputs "ifKernShow";
   if {[winfo exists .kernShow]} {
      kernShowUpdate;
   }
}; # ifKernShow


########################################################################
proc extenShowUpdate {} {
   global tfmpkGlob;
   dputs "extenShowUpdate";
   set w .extenShow;
   if {[winfo exists $w]} {
      set f $tfmpkGlob(extenf);
      set ei [safeCurselection $f.vl];  # ei = exten-entry-index
      dputs "ei=$ei";
      if {$ei >= 0} {
         set extenEntry [$f.vl get $ei];
         scan $extenEntry "%s%s%s%s" v(T) v(M) v(B) v(R);
         set m [sifthenelse [string compare $v(M) ---] trmrb trb];
         $w.i configure -image $tfmpkGlob($m.bitmap);
         
         set msg "";
         foreach X {T M B R} {
            set val $v($X);
            if {[string compare $val ---]} {
               set msg [format "%s%s=%s  " $msg $X $val];
            }
         }
         dputs "extenEntry=$extenEntry msg=$msg";
         set tfmpkGlob(extenMsg) [format $msg];

         set k $tfmpkGlob(extenKVal);
         tfmpk extenShow $ei $k;

         set x $tfmpkGlob(extenXimple);
         set result [tfmpk extenShowWinSize];
         scan $result "%d%d" width height;
         dputs "wh: $width $height";
         set cmd [list $x configure -borderwidth 2 \
                  -width $width -height $height];
         eval $cmd;
         update;
         eval $x redraw; # how will it know  $ei $k ???
      }
   }
}; # extenShowUpdate


########################################################################
proc extenKValPostCtrl {} {
   global tfmpkGlob;
   set state normal;
   if {$tfmpkGlob(extenKVal) == 0} {
      set state disabled;
   }
   .extenShow.fk.down configure -state $state;
   extenShowUpdate;
}; # extenKValPostCtrl


########################################################################
proc extenShowOpen {} {
   global tfmpkGlob;
   dputs "extenShowOpen";
   set w .extenShow;
   if {![winfo exists $w]} {
      topLvlFixedIcon $w "TfmPk Extensible Recipe" Exten;
      pack [label $w.lbl -text "Extensible Recipe"] -side top;
      foreach t {trb trmrb} {
         if {![info exists tfmpkGlob($t.bitmap)]} {
            set absfn [findFile $t.xbm]
            dputs "extenShowOpen: t=$t, absfn=$absfn"
            if {![string compare $absfn ""]} {
               warn "bitmap file $t.xbm not found";
            } else {
               set tfmpkGlob($t.bitmap) [image create bitmap -file $absfn];
               dputs "extenShowOpen: $tfmpkGlob($t.bitmap)";
            }
         }
      }
      pack [label $w.i -image $tfmpkGlob(trb.bitmap)] -side top;
      set tfmpkGlob(extenMsg) "";
      pack [label $w.msg -textvariable tfmpkGlob(extenMsg)] -pady 3;
      pack [frame $w.fx -background black] -side top -padx 5 -pady 5;
      set tfmpkGlob(extenXimple) [ximple $w.fx.extensible];
      pack $tfmpkGlob(extenXimple) -side top -padx 1 -pady 1;

      # k control
      pack [frame $w.fk] -side top -padx 5 -pady 1;
      pack [label $w.fk.klbl -text k=] -side left -padx 2;
      set tfmpkGlob(extenKVal) 0;
      updownVal $w.fk extenKVal extenKValPostCtrl;
      extenKValPostCtrl;

      set bClose [button $w.close -text Close -command [list safeDestroy $w]];
      pack $bClose -side bottom -pady 6;
      extenShowUpdate;
   }
   raise $w;
}; # extenShowOpen


########################################################################
proc pickChar {x y} {
   set pc [tfmpk pickChar $x $y];
   dputs "pickedChar=$pc";
   showFocusChar;
   .tf.f.xft redraw;
}; # pickChar


########################################################################
proc pickChar1 {c} {
   set pc [tfmpk pickChar $c];
   dputs "pickChar1: c=$c, pc=$pc";
   showFocusChar;
   .tf.f.xft redraw;
}; # pickChar1


########################################################################
proc createTableFontWin {} {
   global tfmpkGlob;
   if {![winfo exists .tf]} {
      topLvlFixedIcon .tf "TfmPk Table" Table;
      set tfmpkGlob(tblTitle) "Font Table";
      pack [label .tf.lbl -textvariable tfmpkGlob(tblTitle)] -side top -pady 3;
      pack [frame .tf.f -background black] -padx 3;
      ximple .tf.f.xft -cursor hand1; # font table
      pack .tf.f.xft -padx 2 -pady 2;
      bind .tf.f.xft <ButtonPress> [list pickChar %x %y];
      pack [set fbot [frame .tf.fbot]] -side bottom;
      set msg [label $fbot.msg -textvariable tfmpkGlob(tabMsg) -width 32];
      set cls [button $fbot.close -text Close -command [list safeDestroy .tf]];
      pack $msg $cls -side left -padx 6 -pady 6;
      bind .tf.f.xft <Enter> [list set tfmpkGlob(tabMsg) \
                              "Click to set focus character"];
      bind .tf.f.xft <Leave> [list set tfmpkGlob(tabMsg) ""];
   }
}; # createTableFontWin


########################################################################
proc setSPFlabl {} {
   global tfmpkGlob;
   set tfmpkGlob(SPFstatus) \
      [format "Screen/Font = %d/%d" \
       $tfmpkGlob(fpsNomAct) $tfmpkGlob(fpsDenomAct)];
}; # setSPFlabl


########################################################################
proc tableDraw [] {
   global tfmpkGlob;
   dputs tableDraw;
   set pkfn [basename $tfmpkGlob(currentPK)];
   set tfmfn [basename $tfmpkGlob(currentTFM)];
   set tfmpkGlob(tblTitle)  "<$tfmfn + $pkfn>  -  Table";
   set mg [sifthenelse $tfmpkGlob(maxTabGeo) $tfmpkGlob(geomMaxTab) "0x0"];
   set result [tfmpk tableWinSize $mg $tfmpkGlob(fpsNom) $tfmpkGlob(fpsDenom)];
   dputs "tableWinSize: result=$result";
   # scan $result "%d%d%d%d" width height tfmpkGlob(fpsNom) tfmpkGlob(fpsDenom);
   scan $result "%d%d%d%d" width height \
                           tfmpkGlob(fpsNomAct) tfmpkGlob(fpsDenomAct);
   setSPFlabl;
   .tf.f.xft configure -width $width -height $height;
   updPixelChangers;
   update;
   .tf.f.xft redraw;
}; # tableDraw


########################################################################
proc bringUpTableWin {} {
   global tfmpkGlob;
   if {$tfmpkGlob(dataLoaded)} {
      createTableFontWin;
      wm deiconify .tf;
      raise .tf;
      tableDraw;
   } else {
      warn "No Data loaded";
   }
}; # bringUpTableWin


########################################################################
proc bringUpFocusChar {} {
   global tfmpkGlob;
   if {$tfmpkGlob(dataLoaded)} {
      showFocusChar;
      wm deiconify .focusChar;
      raise .focusChar;
   } else {
      warn "No Data loaded";
   }
}; # bringUpFocusChar


########################################################################
proc setXsizes { plist } {
 global tfmpkGlob; # tfmfont pkfont;
 upvar $plist list;
 dputs "setXsizes Enter";
   set maxNameLength 0;
   set maxTFMPKlength 0;
   set maxComDirlength 0;
   # foreach name $list 
   set nFonts [llength $tfmpkGlob(fontList)];
   for {set i 0} {$i < $nFonts} {incr i} {
      set name [lindex $tfmpkGlob(fontList) $i];
      set l [string length $name];
 dputs "name=$name, l=$l";
      if {$maxNameLength < $l} {
         set maxNameLength $l;
      }

      set tfmfn [lindex $tfmpkGlob(tfmList) $i];
      set pkfn  [lindex $tfmpkGlob(pkList) $i];
      set lcom [comdirlen $pkfn $tfmfn];

      set l [string length $pkfn];
      set l [expr $l-$lcom];
      if {$maxTFMPKlength < $l} {
         set maxTFMPKlength $l;
      }

      set l [string length $tfmfn];
      set l [expr $l-$lcom];
      if {$maxTFMPKlength < $l} {
         set maxTFMPKlength $l;
      }

      if {$maxComDirlength < $lcom} {
         set maxComDirlength $lcom;
      }
   }

 dputs "maxNameLength=$maxNameLength, maxTFMPKlength=$maxTFMPKlength";
 dputs "maxComDirlength=$maxComDirlength";

   .files.comdire configure -width $maxComDirlength;
   .files.tfme    configure -width $maxTFMPKlength;
   .files.pke     configure  -width $maxTFMPKlength;
}; # setXsizes


########################################################################
proc tfmShowDisplay {} {
   global tfmpkGlob;
   dputs "tfmShowDisplay";
   if [winfo exists .tfminfo] {
      foreach param $tfmpkGlob(fwParList) {
         dputs "tfmShowDisplay param=$param";
         set paramf [format "%sf" $param];
         set vl [format "%s.vl" $tfmpkGlob($paramf)];
         dputs "vl=$vl";
         tfmlb $param;
      }
   }

   # active/disactive kern+exten shows
   foreach param {kern exten} {
      set nparams [format "n%ss" $param];
      set n [tfmpk tfminfo $nparams];
      dputs "param=$param, n=$n";
      set b $tfmpkGlob([format "%sShowButton" $param]);
      set w [format ".%sShow" $param];
      if {$n} {
         $b configure -state normal;
         if {[winfo exists $w]} {
            eval [format %sShowWinSize $param];
            eval [format %sShowUpdate $param];
         }
      } else {
         $b configure -state disabled;
         safeDestroy $w;
      }
   }
}; # tfmShowDisplay


########################################################################
proc tfmShow {} {
   global tfmpkGlob;
   foreach param $tfmpkGlob(fwParList) {
      set nparams [format "n%ss" $param];
      dputs "param=$param, nparams=$nparams";
   }
   foreach param $tfmpkGlob(fwParList) {
      dputs "tfmShow param=$param";
      set nparams [format "n%ss" $param];
      set n [tfmpk tfminfo $nparams];
      set Param [capitalFirst $param];
      set nParams [format "n%ss" $Param];
      set tfmpkGlob($nParams) $n;
      set tfmpkGlob([format %ssLabel $param]) [format "%ss\[%s\]" $Param $n];
      set paramsList [format %ssList $param];
      dputs "paramsList=$paramsList"
      set tfmpkGlob($paramsList) {};
      for {set i 0} {$i < $n} {incr i} {
         set wf [tfmpk tfminfo $param $i float];
         # set iwf [format "\[%d\] %s" $i $wf];
         set iwf $wf;
         dputs "i=$i, wf=$wf";
         lappend tfmpkGlob($paramsList) $iwf;
      }
      dputs "list($param)=$tfmpkGlob($paramsList)";
   }
   set tfmpkGlob(tfmGeneral) [tfmpk tfminfo general];
   tfmShowDisplay;
}; # tfmShow


########################################################################
proc pkCharInfo {pc} {
   global tfmpkGlob;
   set pkCharInfo [tfmpk pkCharInfo $pc];  dputs "pkCharInfo=$pkCharInfo";
   scan $pkCharInfo "%d%d%d%d%d%d" dx dy w h hoff voff;
   if {0} {
      set txt [format "Bitmap %dx%d  Offset=(%d,%d) dx=%d" \
              $w $h $hoff $voff $dx];
      set tfmpkGlob(focusPKinfo) $txt;
   }
   set tfmpkGlob(focusPKinfBM) [format "%dx%d" $w $h];
   set tfmpkGlob(focusPKinfHV) [format "+(%d,%d)" $hoff $voff];
   set tfmpkGlob(focusPKinfDX) $dx;
}; # pkCharInfo


########################################################################
proc showFocusChar {} {
   dputs "showFocusChar";
   global tfmpkGlob;
   focusCharWindow;
   set pc [tfmpk pickChar];
   set result [tfmpk focusCharWinSize];
   scan $result "%d%d" width height;

   set cmd [list $tfmpkGlob(winfc) configure \
           -borderwidth 2 -width $width -height $height];
   dputs $cmd
   eval $cmd;
   update;

   set cmd [list $tfmpkGlob(winfc) redraw];
   dputs $cmd
   eval $cmd;
   set tfmpkGlob(fflab) [format "Focus Character: 0x%02x\n" $pc];
   pkCharInfo $pc;

   set result [tfmpk tfminfo charinfo $pc];
   dputs "charinfo=$result";
   scan $result "%d%d%d%d%d%d" \
               i(width) i(height) i(depth) i(italic) tag reminder
   set i(ligkern) -1;
   set i(kern) -1;
   set i(exten) -1;
   set tagText "";
   dputs "tag=$tag";
   if {$tag == 1} { # ligature/kerning
      set i(ligkern) $reminder;
      set lkEntry [lindex $tfmpkGlob(ligkernsList) $i(ligkern)];
      dputs "lkEntry=$lkEntry";
 set llk [string length $lkEntry];
      set kpos [string first "> K " $lkEntry];  # xcc: xcc ff> K nn
 dputs "llk=$llk, kpos=$kpos";
      if {$kpos > 0} {
         set i(kern) [string range $lkEntry [expr $kpos+4] end];
         dputs "i(kern)=$i(kern)";
         set tagText [format " Kern=%d" $i(ligkern)];
      } else {
         set tagText [format " Lig=%d" $i(ligkern)];
      }
      
   }
   if {$tag == 2} { # list
      # puts "Not Supported Info: Ascending characters";
      set tagText [format " list->0x%02x" $reminder];
   }
   if {$tag == 3} { # ext
      set i(exten) $reminder;
      set tagText " Ext=$reminder";
   }
   if [winfo exists .tfminfo] {
      foreach param {width height depth italic ligkern kern exten} {
         set parind $i($param);
         selectParamIndex $param $parind;
      }
   }

   set tfmInfo [format "w=%d h=%d d=%d i=%d%s" \
                $i(width) $i(height) $i(depth) $i(italic) $tagText];
   set tfmpkGlob(focusTFMinfo) $tfmInfo;
   
}; #  showFocusChar


########################################################################
proc show {} {
   global tfmpkGlob;
   dputs "Show";
   createTableFontWin;
   tableDraw;
   update;
   showFocusChar;
   dputs "calling tfmShow";
   tfmShow;
   set tfmpkGlob(pkGeneral) [tfmpk pkGeneral];

}; # show


########################################################################
#                            Preferences                               #
########################################################################

########################################################################
# Get farest 'corner' color from a given color
proc farestColor {someWin c} {
   set rgb [winfo rgb $someWin $c];
   set rgbFar "#";
   set allNone(0) "00";
   set allNone(1) "ff";
   for {set i 0} {$i < 3} {incr i} {
      set c16 [lindex $rgb $i];
      set c8 $allNone([expr $c16 < 0x8000]);
      set rgbFar [format "%s%s" $rgbFar $c8];
      # puts "i=$i, c8=$c8, c16=$c16, rgbFar=$rgbFar";
   }
   return $rgbFar;
}; # farestColor


########################################################################
# preferences defaults
proc prefSetDefaults {} {
   global tfmpkGlob;
   dputs "prefSetDefaults";
   # colors.  See: /usr/X11/lib/X11/rgb.txt
   set tfmpkGlob(backgroundColor) Grey;
   set tfmpkGlob(fontColor)       black;
   set tfmpkGlob(boundaryColor)   brown;
   set tfmpkGlob(textColor)       DarkSlateGrey;
   set tfmpkGlob(pickColor)       DarkGreen;
   set tfmpkGlob(ascendColor)     red;
   # pixel ratios
   set tfmpkGlob(fpsNom) 1;
   set tfmpkGlob(fpsDenom) 1;
   set tfmpkGlob(fpsNomAct) 1;
   set tfmpkGlob(fpsDenomAct) 1;
   set tfmpkGlob(spfZoom) 2;
}; # prefSetDefaults


########################################################################
# Capialize word
proc capitalWord {word} {
   set upc [string toupper [string range $word 0 0]];
   set tail [string range $word 1 end];
   set rc [format "%s%s" $upc $tail];
   return $rc;
}; # capitalWord


########################################################################
proc itemEditColor {item button} {
   global tfmpkGlob;
   dputs "itemEditColor $item $button";
   set currColor [lindex [$button config -background] 4];
   set rgb [winfo rgb $button $currColor];
   set srgb "#";
   for {set i 0} {$i < 3} {incr i} {
      set c8 [expr [lindex $rgb $i]/256];     # 2 significant hex
      set srgb [format "%s%02x" $srgb  $c8];  # append factor
   }
   dputs "currColor=$currColor, srgb=$srgb";
   # set newColor [editColor "TfmPk - Edit Color" $item $srgb];
   set newColor [tk_chooseColor -initialcolor $srgb -parent . \
                   -title "TfmPk - Edit $item Color"];
   if {[string compare $newColor ""]} {
      set tfmpkGlob([format "%sColor" $item]) $newColor;
      $button config -background $newColor

      # Now see if the new color is light or dark
      set rgb [winfo rgb $button $newColor];
      set srgb 0;
      for {set i 0} {$i < 3} {incr i} {
         set c8 [expr [lindex $rgb $i]/256];
         incr srgb $c8;
      }
      dputs "sumRGB = $srgb";
      set sumMidGray [expr 3*0x100/2];
      if {$srgb < $sumMidGray} { # light
         $button config -foreground white;
      } else {
         $button config -foreground black;
      }

      # Now try to generate the farset 'corner' color for the forground-label
      set rgbFgd [farestColor $button $newColor];
      $button config -foreground $rgbFgd;
   }
}


########################################################################
# Activate and deactivate the incr/decr buttons
proc updPixelChangers {} {
   global tfmpkGlob;

   if {[winfo exists .pref]} {

      if {![info exists tfmpkGlob(buttonBoolState,0)]} {
         set tfmpkGlob(buttonBoolState,0) disabled;
         set tfmpkGlob(buttonBoolState,1) normal;
      }

      set b [expr $tfmpkGlob(fpsNom) < $tfmpkGlob(fpsDenom)];
      $tfmpkGlob(fpsNom,butUp) configure \
         -state $tfmpkGlob(buttonBoolState,$b);
      $tfmpkGlob(fpsDenom,butDown) configure \
         -state $tfmpkGlob(buttonBoolState,$b);

      set b [expr $tfmpkGlob(fpsNom) > 1];
      $tfmpkGlob(fpsNom,butDown) configure \
         -state $tfmpkGlob(buttonBoolState,$b);

      set b [expr $tfmpkGlob(spfZoom) > 1];
      $tfmpkGlob(spfZoom,butDown) configure \
         -state $tfmpkGlob(buttonBoolState,$b);
   }
}; # updPixelChangers


########################################################################
proc updTabPixelChangers {} {
   global tfmpkGlob;
   set tfmpkGlob(maxTabGeo) 0;
   updPixelChangers;
}; # updTabPixelChangers


########################################################################
proc updVal {var lab delta postfunc} {
   dputs "updVal $var $lab $delta postfunc=$postfunc";
   global tfmpkGlob;
   incr tfmpkGlob($var) $delta;
   # set v [incr tfmpkGlob($var) $delta];
   # $lab configure -text $v;
   if {[string compare $postfunc ""]} {
      eval $postfunc;
   }
}; # updVal


########################################################################
# Generate a displayed controlled a parameter with incr+decr buttons.
proc updownVal {par var {postfunc ""}} {
   dputs "updownVal par=$par var=$var";
   global tfmpkGlob;
   #set val $tfmpkGlob($var);
   #set lab [label  $par.val -text $val];
   #dputs "updownVal: val=$val";
   set lab [label  $par.val -textvariable tfmpkGlob($var)];
   foreach du {dnarrow uparrow} {
      if {![info exists tfmpkGlob($du.bitmap)]} {
         set absfn [findFile $du.xbm]
         dputs "updownVal: du=$du, absfn=$absfn"
         if {![string compare $absfn ""]} {
            warn "bitmap file $du.xbm not found";
         } else {
            set tfmpkGlob($du.bitmap) [image create bitmap -file $absfn];
         }
      }
   }
   set tfmpkGlob($var,butDown) [
      button $par.down -image $tfmpkGlob(dnarrow.bitmap) -command [
         list updVal $var $lab -1 $postfunc]];
   set tfmpkGlob($var,butUp) [
      button $par.up   -image $tfmpkGlob(uparrow.bitmap) -command [
         list updVal $var $lab 1 $postfunc]];
   pack $par.down -side left
   pack $par.val -side left
   pack $par.up -side left
   set helptag "";
   if {![string compare [string range $var 0 2] fps]} {
      set helptag tablezoom; 
   } elseif {![string compare $var spfZoom]} {
      set helptag focusZoom; 
   }
   if {[string compare $helptag ""]} {
      bind $par.down <Key-F1> [list help $helptag %#];
      bind $par.up <Key-F1> [list help $helptag %#];
   }

}; # updownVal


########################################################################
proc prefResetDefaults {clrBut listFrameVar} {
   global tfmpkGlob;
   prefSetDefaults;

   foreach item {background font boundary text pick ascend} {
      dputs "prefResetDefaults item=$item";
      set b [format "%s.%s" $clrBut $item];
      set bgdColor $tfmpkGlob([format "%sColor" $item]);
      set fgdColor [farestColor $clrBut $bgdColor];
      $b configure -background $bgdColor -foreground $fgdColor;
   }

   foreach frameVar $listFrameVar {
      dputs "frameVar=$frameVar";
      set par [lindex $frameVar 0];
      set var [lindex $frameVar 1];
      set val $tfmpkGlob($var);
      set val2 [expr $val*$val];
      set lab $par.val;
      $lab configure -text $val2;
   }
   updPixelChangers;
}; # prefResetDefaults


########################################################################
proc setPrefs {} {
   global tfmpkGlob;

   # colors
   foreach item {background font boundary text pick ascend} {
      dputs "setPrefs item=$item";
      set bgdColor $tfmpkGlob([format "%sColor" $item]);
      tfmpk pref color $item $bgdColor;
   }

   # check valid geometry
   set r2 [scan $tfmpkGlob(geomMaxTab) "%dx%d" mX mY];
   if {$r2 != 2} {
      warn "Bad (maxtab) geometry specification in $tfmpkGlob(geomMaxTab)";
   } else {
      set tfmpkGlob(geomMaxTabLast) [format "%dx%d" $mX $mY]; 
   }
   set tfmpkGlob(geomMaxTab) $tfmpkGlob(geomMaxTabLast);

   foreach item {fpsNom fpsDenom spfZoom} {
      tfmpk pref pixel $item $tfmpkGlob($item);
   }
}; # setPrefs


########################################################################
proc prefApply {} {
   global tfmpkGlob;
   setPrefs;
   if {$tfmpkGlob(dataLoaded)} {
      show;
   }
}; # prefApply


########################################################################
proc prefOK {} {
   prefApply;
   safeDestroy .pref;
}; # prefOK


########################################################################
proc editPref {} {
   global tfmpkGlob;
   dputs "editPref";
   if {![winfo exists .pref]} {
      dputs "Create .pref";
      topLvlFixedIcon .pref "TfmPk - Preferences" Pref;
      pack [label .pref.title -text Preferences] -side top -pady 3;

      set clrBut [frame .pref.clrbut -bd 3];
      set colorLabel [label $clrBut.l -text Colors:];
      pack $colorLabel -side top;

      foreach item {background font boundary text pick ascend} {
         dputs "item=$item";
         set b [format "%s.%s" $clrBut $item];
         set bgdColor $tfmpkGlob([format "%sColor" $item]);
         set fgdColor [farestColor $clrBut $bgdColor];
         button $b -text [capitalWord $item] \
            -background $bgdColor -foreground $fgdColor \
            -command [list itemEditColor $item $b];
         pack $b -fill x;
      }

      ########################################################################
      #                            Pixels ratio                              #
      ########################################################################

      set pixFrame [frame .pref.pix -bd 3];
      pack [label $pixFrame.title -text "Pixel Ratios"] -side top;

      #####################################
      # Font-pixels per Screen pixels ratio

      set fps [frame $pixFrame.fps];
      pack $fps -side top -pady 3 ;

      pack [set fmg [frame $fps.fmg]] -side top
         pack [label $fmg.l -text "Max Table Geometry"] -side top -pady 5;
         set cb [checkbutton $fmg.ch -variable tfmpkGlob(maxTabGeo)];
         pack $cb -side left;
         pack [entry $fmg.e -width 9 -textvariable tfmpkGlob(geomMaxTab)] \
            -side left;
         bind $cb <Key-F1> [list help maxgeom %#];
         bind $fmg.e <Key-F1> [list help maxgeom %#];

      # label $fps.l -text Font/Screen:
      setSPFlabl;
      label $fps.l -textvariable tfmpkGlob(SPFstatus);
      pack $fps.l -side top
      set fpsnom [frame $fps.nom];
      set fpsdenom [frame $fps.denom];
      pack $fpsnom -side top;
      pack $fpsdenom -side top;

      updownVal $fpsnom fpsNom updTabPixelChangers;      # nominator
      updownVal $fpsdenom fpsDenom updTabPixelChangers;  # denominator
      bind $fps <Key-F1> [list help tablezoom %#];
      #
      #####################################

      ###############################################
      # Zoom character, Screen-pixels per Font-pixels
      set spfzoom [frame $pixFrame.zoom];
      pack $spfzoom -pady 3;

      label $spfzoom.l -text Focus:
      pack $spfzoom.l

      updownVal $spfzoom spfZoom updPixelChangers;
      bind $spfzoom <Key-F1> [list help focusZoom %#];
      #
      ###############################################
      updPixelChangers;

      # Bottom
      set low [frame .pref.low -bd 10];
      if (1) {
         button $low.apply -text Apply -command prefApply;
         set listFrameVar [list [list $fpsnom fpsNom] \
                                [list $fpsdenom fpsDenom] \
                                [list $spfzoom spfZoom]];
         button $low.ok -text OK -width 5 -command prefOK;
         button $low.defaults -text Reset \
            -command [list prefResetDefaults $clrBut $listFrameVar];
         button $low.dismiss -text Close -command [list safeDestroy .pref];
         pack $low.apply -side left -anchor center;
         pack $low.ok -side left -anchor center;
         pack $low.defaults -side left -anchor center;
         pack $low.dismiss -side right -anchor center;
      }

      # Main Packing
      pack $low -side bottom -fill x;
      pack $pixFrame -side right -anchor n -fill y;
      pack $clrBut -side left -padx 3;

      bind .pref <Key-F1> [list help pref %#];
   }

   wm deiconify .pref;
   raise .pref;
}; # editPref


########################################################################
proc usage {} {
   puts "Usage:";
   puts "tfmpk \[-tfmpath <TFMpath>] \[-pkpath <PKpath>]";
   puts "      \[-load <file.tfm> <file.nnnpk>]";
   puts "      \[-bitmapicon]";
   puts "      \[-color <item> <color>]";
   puts "      \[-fontPerScreen <num> <denom>]";
   puts "      \[-zoom <n>]";
   puts "      \[-tabmaxgeo <<width>x<hight>>]";
   puts "      \[-flags <uf>]";
   puts "      \[-help]";
   puts "";
   puts "Where:";
   puts "      item is one of {background font boundary text pick ascend}";
   puts "      uf is a bitwise combination of:";
   puts "         1 = Warn about tfm-file size";
}; # usage


########################################################################
# Search path for basic tfm+pk file
proc findLoadTfmPk {tfmBase pkBase} {
   global tfmpkGlob;
   set etfm "";
   set epk "";
   set e(tfm) $tfmBase;
   set e(pk) $pkBase;
   set ok 1; 
   foreach tp {tfm pk} {
      if {$ok && ![file exist $e($tp)]} {
         set path $tfmpkGlob([format %sPath $tp]);  
         dputs "etp=$e($tp), path=$path";
         set l [lspath -base $e($tp) $path];
         if {[llength $l] > 0} {
            set e($tp) [lindex $l 0];
         } else {
            warn "Could not find $e($tp)";
            set e($tp) "";
            set ok 0;
         }
      }
   }
   if {$ok} {
      set tfmpkGlob(etfm) $e(tfm)
      set tfmpkGlob(epk) $e(pk)
      loadTfmPk;
   }
}; # findLoadTfmPk


########################################################################
# Program main Tcl-routine
proc tfmpkMain {args} {
   global debug tfmpkGlob tk_strictMotif;

   dputs "tfmpkMain: args=$args";
   set tk_strictMotif 0; # I want Emacs-editing style. but see selectTFMPK

   # set arg0  for sarching files
   set tfmpkGlob(arg0) [lindex $args 0];

   # The C-main program wants to save command-line arguments
   eval tfmpk init $args;

   tfmpk winit;

   initGlobals;

   # Initialize Prefereces !!
   prefSetDefaults;

   # First, let's set the default tfm and pk path before the user messes them.
   setFontPaths;

   # Finally command line argument override.
   set nullstr "";
   set argc [llength $args];    dputs "argc=$argc, args=$args";
   set i 1;
   set doLoad 0;
   while {$i < $argc} {
      set argvi [lindex $args $i];    dputs "i=$i, argvi=$argvi";
      set i1 [expr $i + 1];
      set argvi1 [lindex $args $i1];
      set i2 [expr $i1 + 1];
      set argvi2 [lindex $args $i2];
      if {![string compare $argvi "-help"]} {
         usage;
         set tfmpkGlob(initscript) help; # tellmain.c to exit
         set i $argc;
      } elseif {![string compare $argvi "-tfmpath"]} {
         set tfmpkGlob(tfmPath) $argvi1;
         set i [expr $i+2];
      } elseif {![string compare $argvi "-pkpath"]} {
         set tfmpkGlob(pkPath) $argvi1;
         set i [expr $i+2]
      } elseif {![string compare $argvi "-load"]} {
         set doLoad 1;
         set etfm $argvi1;
         incr i1;
         set epk $argvi2;
         set i [expr $i+3]
      } elseif {![string compare $argvi "-bitmapicon"]} {
         dputs bitmapicon;
         set tfmpkGlob(bitmapIcon) 1;
         set i [expr $i+1]
      } elseif {![string compare $argvi "-color"]} {
         set item [format %sColor $argvi1];
         set tfmpkGlob($item) $argvi2;
         set i [expr $i+3]
      } elseif {![string compare $argvi "-flags"]} {
	 set tfmpkGlob(flags) $argvi1;
         set i [expr $i+2]
      } elseif {![string compare $argvi "-fontPerScreen"]} {
         set tfmpkGlob(fpsNom) $argvi1;
         set tfmpkGlob(fpsDenom) $argvi2;
         set i [expr $i+3]
      } elseif {![string compare $argvi "-zoom"]} {
         set tfmpkGlob(spfZoom) $argvi1;
         set i [expr $i+2];
      } elseif {![string compare $argvi "-tabmaxgeo"]} {
         set tfmpkGlob(geomMaxTab) $argvi1;
         set i [expr $i+2];
      } else {
         set i $argc;
      }
   }
   dputs "args scanned";
   mainWindow;


   if {([string length $tfmpkGlob(tfmPath)] == 0) || \
       ([string length $tfmpkGlob(pkPath)] == 0)} {
      usage;
      exit 1;
   }

   setPrefs;

   if {$debug} {
     puts "tfmpk using:"
     puts "  tfmPath = $tfmpkGlob(tfmPath)"
     puts "  pkPath = $tfmpkGlob(pkPath)"
   }

   if {$doLoad} {
      findLoadTfmPk $etfm $epk;
   }
}; # tfmpkMain


########################################################################
proc help {item serial {parent .}} {
   global tfmpkGlob;
   set s "Unknown Help Item=$item";
   dputs "help item=$item serial=$serial";
   if {$tfmpkGlob(serial) != $serial} {
      if {$serial != -1} {
         set tfmpkGlob(serial) $serial;
      }
   
      if {![string compare $item general]} {
         set s [format "%s%s%s%s%s" \
         "TfmPk is a fonts-viewer for TeX. " \
         "Each font comes in pairs of files. " \
         "Font Metric: foo.tfm and Pixel compressed map: foo.<nnn>pk.\n\n" \
         "Use the <F1> Key for context-help.\n" \
         "Use <Tab> to change keyboard Focus."];
      } elseif {![string compare $item loadFont]} {
         set s [format %s%s \
            "The load Font menu helps to select and load " \
            "a pair of .tfm & .pk files."];
      } elseif {![string compare $item NameRegExp]} {
         set s [format %s%s%s \
            "The Name Regular Expression restricts the listing of fonts " \
            "for next (Re)List command. For example: to restrict listing " \
            " only to fonts whose name starts with \"cmr\", use  ^cmr."];
      } elseif {![string compare $item DPIRegExp]} {
         set s [format %s%s%s%s%s \
            "You can restrict the listing of fonts to specfic " \
            "resolutions (Dots Per Inch) by specifying " \
            "DPI Regular Expression.\n" \
            "For example to list only 300pk or 360pk, " \
            "use 300|360."];
      } elseif {![string compare $item relist]} {
         set s [format %s%s%s%s \
            "Scan the directories in the TFM Path and PK Path. " \
            "Filter with Name+DPI Regular Expression, and match names " \
            "of foo.<nnn>pk files with foo.tfm.\n" \
            "Finally list the result. "];
      } elseif {![string compare $item listselect]} {
         set s [format %s%s \
            "Click to set the tfm-file & pk-file entries.\n " \
            "Double Click also loads."];
      } elseif {![string compare $item pref]} {
         set s "Set preferences for colors and zoom ratios";
      } elseif {![string compare $item maxgeom]} {
         set s [format %s%s \
            "Set maximal geometry in pixels for the font table.\n" \
            "If Max is set, then Screen/Font ratio numbers are ignored."]
      } elseif {![string compare $item tablezoom]} {
         set s [format %s%s \
            "Set the reduction size of the font table as a ratio of:\n" \
            "Screen-Pixel / Font-Bitmap-Pixel\n"];
      } elseif {![string compare $item focusZoom]} {
         set s "Set the magnification multiplier of the focus character";
      } elseif {![string compare $item notYet]} {
      }
      tk_messageBox -title "TfmPk Help" \
         -parent $parent -icon info -type ok -message $s;
   }
}
# exit 0



