// $Id: tk.c,v 1.2 1997/03/25 10:18:43 yotam Exp $
// Simple wrappers enhancing Tk

#include <string.h>
#include <X11/Xlib.h>
#include <tk.h>
#include "baseint.h"
#include "block.h"
#include "tkinfo.h"
#include "debug.h"


#if DEBUG
#include <iostream.h>
////////////////////////////////////////////////////////////////////////
ostream&  operator << (ostream& s, const XWindowAttributes& wa)
{
   s << wa.width << 'x' << wa.height << '+' << wa.x << '+' << wa.y << 
        " bs=" << wa.backing_store << ", save_under=" << wa.save_under <<
        ", map_state=" << wa.map_state << endl;
   return(s);
} // ostream << XWindowAttributes&
#endif


////////////////////////////////////////////////////////////////////////
TkInfo::TkInfo()
{
   DPRNL("TkInfo");
}


////////////////////////////////////////////////////////////////////////
TkInfo::~TkInfo()
{
   DPRNL("~TkInfo");
}


////////////////////////////////////////////////////////////////////////
extern int
tfmpkTkResizeCheck(Tk_Window tkwin, US32 width, US32 height)
{
   int                rc = 0;
   XWindowAttributes  wa;

   Tk_ResizeWindow(tkwin, width, height);
   Tk_MapWindow(tkwin);
   Display  *dpy = Tk_Display(tkwin);
   Window   win  = Tk_WindowId(tkwin);
   if (XGetWindowAttributes(dpy, win, &wa))
   {
      rc = (width == (US32)wa.width) && (height == (US32)wa.height);
   }
   return(rc);
} // tfmpkTkResizeCheck


////////////////////////////////////////////////////////////////////////
static bool
strMatchTail(const char *s, const char* tail)
{
   int   ls = strlen(s);
   int   lt = strlen(tail);
   bool  match = (lt <= ls);
   if (match)
   {
      const char*  stail = s + (ls - lt);
      match = (strcmp(stail, tail) == 0);
   }
   return(match);
} // strMatchTail


////////////////////////////////////////////////////////////////////////
int
TkInfo::assignFuncs
(             // output
   ostream&                     result,
   XimpleDrawFunc*&             drawFunc,
   XimpleCustomConfigFunc*&     custConfFunc,
   XimpleDestroyFunc*&          , // destroyFunc,
   ClientData&                  appData,
              // input
   Tcl_Interp*                  , // interp, 
   Tk_Window                    tkwin,
   int                          , // argc,
   char**                       argv,
   ClientData                   assignData
)
{
   int  rc = TCL_OK;
   TfmpkProgBlock*  pb = (TfmpkProgBlock *)assignData;
   if (strMatchTail(argv[1], ".xft"))
   {
      drawFunc              = drawFontTable;
      custConfFunc          = configureFontTable;
      appData               = assignData;
      pb->pTk->fontTableWin = tkwin;
   }
   else if (strMatchTail(argv[1], ".focusChar"))
   {
      drawFunc              = drawFocusChar;
      custConfFunc          = configureFocusChar;
      appData               = assignData;
      pb->pTk->winFocusChar = tkwin;
   }
   else if (strMatchTail(argv[1], ".kern"))
   {
      drawFunc              = kernDraw;
      custConfFunc          = kernConfigure;
      appData               = assignData;
      pb->pTk->kernWin    = tkwin;
   }
   else if (strMatchTail(argv[1], ".extensible"))
   {
      drawFunc              = extenDraw;
      custConfFunc          = extenConfigure;
      appData               = assignData;
      pb->pTk->extenWin    = tkwin;
   }
   else
   {
      result << "Could not match ximple-widget with " << argv[1] << ends;
      rc = TCL_ERROR;
   }
   return(rc);
} // assignFuncs

