// $Id: ximple.c,v 1.3 1997/04/01 08:21:43 yotam Exp $

#include <strstream.h>
#include "ximple.h"
#include "debug.h"

// look at Brent Welch's ~/TKBOOK/examples/tkWidget.c

class Ximple
{
 public:
   Ximple();
   static int                init(Tcl_Interp*           interp,
                                  XimpleAssignFuncs*    vassignFuncs,
                                  ClientData            vassignData,
                                  const Tk_ConfigSpec*  customConfigSpec);
   virtual ~Ximple() {};
 private:
   static int                sCmd(ClientData clientData,
                                  Tcl_Interp* interp,
                                  int         argc,
                                  char**      argv
                                  );
   int                       cmd(ostream      &result,
                                 ClientData   clientData,
                                 Tcl_Interp*  interp,
                                 int          argc,
                                 char**       argv);

   static int                sInstanceCmd(ClientData  clientData,
                                          Tcl_Interp* interp,
                                          int         argc,
                                          char**      argv
                                          );
   int                       instanceCmd(int argc, char** argv);

   int                       configure(Tcl_Interp* interp,
                                       int argc, char** argv, int flags);

   static void               sEventProc(ClientData clientData, XEvent* event);
   void                      eventProc(XEvent* event);

   static void               sCmdDelete(ClientData clientData);

   static void               sDisplay(ClientData clientData);
   void                      display();

   static void               sDestroy(ClientData clientData);
   void                      destroy();

   void                      orderUpdate();
   // End of methods
   /////////////////

   /////////////////
   // data memebers

   static XimpleAssignFuncs*  assignFuncs;
   static ClientData          assignData;

   enum   {MaxConfigSpec = 128};
   static Tk_ConfigSpec       configSpec[MaxConfigSpec];

   Tcl_Interp*                theInterp;
   Tk_Window                  tkwin;
   XimpleDrawFunc*            drawFunc;
   XimpleCustomConfigFunc*    configFunc;
   XimpleDestroyFunc*         destroyFunc;
   ClientData                 appData;
   bool                       updatePending;

   // configuration (canvas-inspired)

   /*
    * Information used when displaying widget:
    */

   int borderWidth;		/* Width of 3-D border around window. */
   Tk_3DBorder bgBorder;	/* Used for canvas background. */
   int relief;			/* Indicates whether window as a whole is
                                 * raised, sunken, or flat. */
   int highlightWidth;		/* Width in pixels of highlight to draw
                                 * around widget when it has the focus.
                                 * <= 0 means don't draw a highlight. */
   int width, height;		/* Dimensions to request for canvas window,
				 * specified in pixels. */
   Cursor cursor;		/* Current cursor for window, or None. */

}; // Ximple


XimpleAssignFuncs*  Ximple::assignFuncs;
ClientData         Ximple::assignData;

Tk_ConfigSpec      Ximple::configSpec[MaxConfigSpec] =
{
   {TK_CONFIG_BORDER, "-background", "background", "Background",
       "#d9d9d9", Tk_Offset(Ximple, bgBorder),
       TK_CONFIG_COLOR_ONLY},
   {TK_CONFIG_BORDER, "-background", "background", "Background",
       "White", Tk_Offset(Ximple, bgBorder),
       TK_CONFIG_MONO_ONLY},
   {TK_CONFIG_SYNONYM, "-bd", "borderWidth", (char *) NULL,
       (char *) NULL, 0, 0},
   {TK_CONFIG_SYNONYM, "-bg", "background", (char *) NULL,
       (char *) NULL, 0, 0},
   {TK_CONFIG_PIXELS, "-borderwidth", "borderWidth", "BorderWidth",
       "0", Tk_Offset(Ximple, borderWidth), 0},
   {TK_CONFIG_ACTIVE_CURSOR, "-cursor", "cursor", "Cursor",
       "", Tk_Offset(Ximple, cursor), TK_CONFIG_NULL_OK},
   {TK_CONFIG_PIXELS, "-height", "height", "Height",
       "7c", Tk_Offset(Ximple, height), 0},
   {TK_CONFIG_RELIEF, "-relief", "relief", "Relief",
       "flat", Tk_Offset(Ximple, relief), 0},
   {TK_CONFIG_PIXELS, "-width", "width", "Width",
       "10c", Tk_Offset(Ximple, width), 0},
   {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
       (char *) NULL, 0, 0}

   // leave room for custom configuration specification
};


////////////////////////////////////////////////////////////////////////
Ximple::Ximple() :
   theInterp(0),
   tkwin(0),
   drawFunc(0),
   configFunc(0),
   destroyFunc(0),
   updatePending(0),
   borderWidth(0),
   bgBorder(0),
   width(0),
   height(0),
   cursor(0)
{
} // Ximple::Ximple


////////////////////////////////////////////////////////////////////////
int
Ximple::sCmd(ClientData clientData, Tcl_Interp* interp, int argc, char** argv)
{

   int          rc = TCL_OK;
   ostrstream   result(interp->result, TCL_RESULT_SIZE - 1);
   Ximple*       ximple = new Ximple;
   if (ximple == 0)
   {
      result << "no memory" << ends;
      rc = TCL_ERROR;
   }
   else
   {
      rc = ximple->cmd(result, clientData, interp, argc, argv);
      if (rc != TCL_OK)
      {
         delete ximple;
      }
   }
   return(rc);
} // Ximple::sCmd



////////////////////////////////////////////////////////////////////////
int
Ximple::cmd
(
   ostream     &result,
   ClientData  clientData,
   Tcl_Interp*  interp,
   int         argc,
   char**        argv
)
{
   int          rc = TCL_OK;
   Tk_Window    newWin;
   if (argc < 2)
   {
      result << "wrong # args: should be \"" <<
                argv[0] << " pathName ?options?\"" << ends;
      rc = TCL_ERROR;
   }
   else
   {
      Tk_Window  mainWin = (Tk_Window)clientData;
      newWin = Tk_CreateWindowFromPath(interp, mainWin, argv[1], 0);
      if (newWin == 0)
      {
         result << "window creation failed" << ends;
         rc = TCL_ERROR;
      }
      else
      {
         rc = (*assignFuncs)(// output
                             result, drawFunc, configFunc, destroyFunc,
                             appData,
                             // input
                             interp, newWin, argc, argv, assignData);
      }
   }

   if (rc == TCL_OK)
   {
      Tk_SetClass(newWin, "Ximple");
      theInterp = interp;
      tkwin = newWin;
      Tk_CreateEventHandler(newWin, ExposureMask | StructureNotifyMask,
                            sEventProc, this);
      Tcl_CreateCommand(interp, Tk_PathName(newWin), sInstanceCmd, this, 
                        sCmdDelete);
      rc = configure(interp, argc - 2, argv + 2, 0);
      if (rc != TCL_OK)
      {
         Tk_DestroyWindow(newWin);
      }
      else
      {
         result << Tk_PathName(newWin) << ends;
      }
   }

   return(rc);
} // Ximple::cmd


////////////////////////////////////////////////////////////////////////
int
Ximple::sInstanceCmd
(
   ClientData   clientData,
   Tcl_Interp*  DCODE(interp),
   int          argc,
   char**       argv
)
{
   Tk_Preserve(clientData);
   Ximple*  ximple = (Ximple *)clientData;
   DASSERT(interp == ximple->theInterp);
   int  rc = ximple->instanceCmd(argc, argv);
   Tk_Release(clientData);
   return(rc);
} // Ximple::sInstanceCmd


////////////////////////////////////////////////////////////////////////
int
Ximple::instanceCmd(int  argc, char**  argv)
{
   int          rc = TCL_OK;
   ostrstream   result(theInterp->result, TCL_RESULT_SIZE - 1);
   if (argc < 2)
   {
      result << "wrong # args: should be \"" <<
                argv[0] << " option ?arg arg ...?\"" << ends;
      rc = TCL_ERROR;
   }
   else if (strcmp(argv[1], "redraw") == 0)
   {
      if (argc != 2)
      {
         result << "redraw must not have parameters" << ends;
         rc = TCL_ERROR;
      }
      else
      {
         orderUpdate();
      }
   }
   else if (strcmp(argv[1], "cget") == 0)
   {
      if (argc != 3)
      {
         result << "wrong # args: should be \""  <<
                    argv[0] << " cget option\"" << ends;
         rc = TCL_ERROR;
      }
      else
      {
         rc = Tk_ConfigureValue(theInterp, tkwin, configSpec,
                                    (char *)this, argv[2], 0);
      }
   }
   else if (strcmp(argv[1], "appconfigure") == 0)
   {
      if (argc != 2)
      {
         result << "\"appconfigure\" should not have more parameters" << endl;
         rc = TCL_ERROR;
      }
      else if (configFunc)
      {
         rc = (*configFunc)(tkwin, appData);
      }
   }
   else if (strcmp(argv[1], "configure") == 0)
   {
      if (argc <= 3)
      { // info on all attributes or one specified
         rc = Tk_ConfigureInfo(theInterp, tkwin, configSpec, (char *)this,
                               (argc == 3 ?  argv[2] : (char *)0),
                                0);
      }
      else // change
      {
         rc = configure(theInterp, argc - 2, argv + 2, TK_CONFIG_ARGV_ONLY);
      }
   }
   else
   {
      result << "bad option \"" << argv[1] << "\"" << ends;
      rc = TCL_ERROR;
   }
   return(rc);
} // Ximple::instanceCmd


////////////////////////////////////////////////////////////////////////
int
Ximple::configure(Tcl_Interp* interp, int argc, char** argv, int flags)
{
   DASSERT(interp == theInterp);
   int  rc = TCL_OK;
   if (1 || argc)
   {
      rc = Tk_ConfigureWidget(interp, tkwin, configSpec, argc, argv,
                              (char *)this, flags);
   }
   if (rc == TCL_OK)
   {
      XColor*  xc = Tk_3DBorderColor(bgBorder);
      Tk_SetWindowBackground(tkwin, xc->pixel);
      Tk_GeometryRequest(tkwin, width + 2*borderWidth, height + 2*borderWidth);
      Tk_SetInternalBorder(tkwin, borderWidth);
      if (configFunc)
      {
         rc = (*configFunc)(tkwin, appData);
         if (rc && Tk_IsMapped(tkwin))
         {
            orderUpdate();
         }
      }
   }
   return(rc);
} // Ximple::configure


////////////////////////////////////////////////////////////////////////
void
Ximple::sEventProc(ClientData clientData, XEvent* event)
{
   Ximple*  ximple = (Ximple *)clientData;
   ximple->eventProc(event);
   if (event->type == DestroyNotify)
   {
      sDestroy((ClientData)ximple);
   }
} // Ximple::sEventProc


////////////////////////////////////////////////////////////////////////
void
Ximple::eventProc(XEvent* event)
{
   if (event->type != DestroyNotify)
   {
      orderUpdate();
   }
} // Ximple::eventProc


////////////////////////////////////////////////////////////////////////
void
Ximple::sDisplay(ClientData clientData)
{
   DPRNL("sDisplay");
   Ximple*  ximple = (Ximple *)clientData;
   ximple->display();
} // Ximple::sDisplay


////////////////////////////////////////////////////////////////////////
void
Ximple::display()
{
   (*drawFunc)(tkwin, appData);
   updatePending = false;
} // Ximple::display


////////////////////////////////////////////////////////////////////////
void
Ximple::sCmdDelete(ClientData DCODE( clientData))
{
   DCODE(Ximple*  ximple = (Ximple *)clientData);
} // Ximple::sCmdDelete


////////////////////////////////////////////////////////////////////////
void
Ximple::sDestroy(ClientData clientData)
{
   Ximple*  ximple = (Ximple *)clientData;
   ximple->destroy();
   delete ximple;
} // Ximple::sDestroy


////////////////////////////////////////////////////////////////////////
void
Ximple::destroy()
{
   if (updatePending)
   {
      Tcl_CancelIdleCall(sDisplay, this);
      // updatePending = false; 
   }
   if (destroyFunc)
   {
      (*destroyFunc)(tkwin, appData);
   }
   Tcl_DeleteCommand(theInterp, Tk_PathName(tkwin));
   Tk_FreeOptions(configSpec, (char *)this, Tk_Display(tkwin), 0);
} // Ximple::Ximple::destroy


////////////////////////////////////////////////////////////////////////
void
Ximple::orderUpdate()
{
   DPR("Ximple::orderUpdate: updatePending " << updatePending << endl);
   if (!updatePending)
   {
      updatePending = true;
      Tcl_DoWhenIdle(sDisplay, this);
   }
} // Ximple::orderUpdate


////////////////////////////////////////////////////////////////////////
int
Ximple::init
(
   Tcl_Interp*           interp,
   XimpleAssignFuncs*    vassignFuncs,
   ClientData           vassignData,
   const Tk_ConfigSpec*  custConfSpec
)
{
   int  rc = TCL_OK;
   assignFuncs = vassignFuncs;
   assignData  = vassignData;
   Tcl_CreateCommand(interp, "ximple", Ximple::sCmd,
                     (ClientData)Tk_MainWindow(interp), 0);
   if (custConfSpec) // append standard specs
   {
      unsigned int  s, t;  // source -> target
      for (t = 0;  configSpec[t].type != TK_CONFIG_END;  t++); // find old end
      for (s = 0;  custConfSpec[s].type != TK_CONFIG_END;  s++); // user end
      if (s + t + 1 >= MaxConfigSpec)
      {
         ostrstream  result(interp->result, TCL_RESULT_SIZE-1);
         result << "Too many custom configuration specifications" << ends;
         rc = TCL_ERROR;
      }
      else
      {
         (void)memcpy(configSpec + t, custConfSpec,
                      s * sizeof(Tk_ConfigSpec));
      }
   }
   return(rc);
} // Ximple::init


////////////////////////////////////////////////////////////////////////
int
Ximple_Init
(
   Tcl_Interp*           interp,
   XimpleAssignFuncs    assignFuncs,
   ClientData           assignData,
   const Tk_ConfigSpec*  customConfigSpec
)
{
   int  rc = Ximple::init(interp, assignFuncs, assignData, customConfigSpec);
   return(rc);
} // Ximple_Init
