// $Id: xwin.c,v 1.2 1997/07/25 05:44:30 yotam Exp yotam $
#include "xwin.h"
#include "FocusCharPixels.h"
#include "debug.h"

////////////////////////////////////////////////////////////////////////
const char* 
XWinInfo::ColorOptionName[XWinInfo::Ngc] = 
{
   "backround",
   "forground",
   "borderColor",
   "textColor",
   "selectColor",
   "ascendColor"
};


////////////////////////////////////////////////////////////////////////
XWinInfo::XWinInfo(Tcl_Interp*  interp) :
   _interp(interp),
   _focusCharPixels(new FocusCharPixels)
{
   _someWin = Tk_MainWindow(interp);
   _dpy = Tk_Display(_someWin);
   Window  wid = Tk_WindowId(_someWin);
   for (int i = Ngc;  i--;)
   {
      const char* colorName = Tk_GetOption(_someWin,
                                           (char *)ColorOptionName[i],
                                           (char *)0);
      DPRNL("colorOption=" << ColorOptionName[i] << ", name=" << colorName);
      if (colorName == 0)
      {
         colorName = (i == XWinInfo::Clear ? "white" : "black");
      }
      _colorSpec[i] = string(colorName);
      _xColor[i] = Tk_GetColor(interp, _someWin, (char*)colorName);
      _gc[i] = Tk_GCForColor(_xColor[i], wid);
   }
} // XWinInfo::XWinInfo


////////////////////////////////////////////////////////////////////////
XWinInfo::~XWinInfo()
{
   int  i;
   DASSERT(_dpy != 0);
   for (i = Ngc;  i--;)
   {
      Tk_FreeColor(_xColor[i]);
   }
   if (_focusCharPixels)
   {
      delete _focusCharPixels;
   }
} // XWinInfo::~XWinInfo


////////////////////////////////////////////////////////////////////////
bool
XWinInfo::setColorGC(const char* item, const char* colorSpec)
{
   static const char* itemName[Ngc] = 
                     {"background", 
                      "font", 
                      "boundary", 
                      "text", 
                      "pick", 
                      "ascend"
                      };
   bool match;
   int  i;
   for (i = Ngc, match = false;  !match && i--;  )
   {  DASSERT(itemName[i] && itemName[i][0]);
      match = (strcmp(item, itemName[i]) == 0);
   }
   DASSERT(match);
   if (match)
   {
      string  newColorSpec(colorSpec);
      if (_colorSpec[i] != newColorSpec)
      {
         _colorSpec[i] = newColorSpec;
         DASSERT(_xColor[i] != 0);
         Tk_FreeColor(_xColor[i]);
         _xColor[i] = Tk_GetColor(_interp, _someWin, (char*)colorSpec);
         _gc[i] = Tk_GCForColor(_xColor[i], Tk_WindowId(_someWin));
      }
   }
   return(match);
} // setColorGC
